"""
This is an implementation of the Linear Fascicle Evaluation (LiFE) algorithm
described in :footcite:p:`Pestilli2014`.

References
----------
.. footbibliography::
"""

import numpy as np
import scipy.linalg as la
import scipy.sparse as sps

import dipy.core.optimize as opt
import dipy.data as dpd
from dipy.reconst.base import ReconstFit, ReconstModel
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import transform_streamlines
from dipy.tracking.utils import unique_rows
from dipy.tracking.vox2track import _voxel2streamline


def gradient(f):
    """
    Return the gradient of an N-dimensional array.

    The gradient is computed using central differences in the interior
    and first differences at the boundaries. The returned gradient hence has
    the same shape as the input array.

    Parameters
    ----------
    f : array_like
      An N-dimensional array containing samples of a scalar function.

    Returns
    -------
    gradient : ndarray
      N arrays of the same shape as `f` giving the derivative of `f` with
      respect to each dimension.

    Examples
    --------
    >>> x = np.array([1, 2, 4, 7, 11, 16], dtype=float)
    >>> gradient(x)
    array([ 1. ,  1.5,  2.5,  3.5,  4.5,  5. ])

    >>> gradient(np.array([[1, 2, 6], [3, 4, 5]], dtype=float))
    [array([[ 2.,  2., -1.],
           [ 2.,  2., -1.]]), array([[ 1. ,  2.5,  4. ],
           [ 1. ,  1. ,  1. ]])]

    Notes
    -----
    This is a simplified implementation of gradient that is part of numpy
    1.8. In order to mitigate the effects of changes added to this
    implementation in version 1.9 of numpy, we include this implementation
    here.
    """
    f = np.asanyarray(f)
    N = len(f.shape)  # number of dimensions
    dx = [1.0] * N

    # use central differences on interior and first differences on endpoints
    outvals = []

    # create slice objects --- initially all are [:, :, ..., :]
    slice1 = [slice(None)] * N
    slice2 = [slice(None)] * N
    slice3 = [slice(None)] * N

    for axis in range(N):
        # select out appropriate parts for this dimension
        out = np.empty_like(f)
        slice1[axis] = slice(1, -1)
        slice2[axis] = slice(2, None)
        slice3[axis] = slice(None, -2)

        # 1D equivalent -- out[1:-1] = (f[2:] - f[:-2])/2.0
        out[tuple(slice1)] = (f[tuple(slice2)] - f[tuple(slice3)]) / 2.0
        slice1[axis] = 0
        slice2[axis] = 1
        slice3[axis] = 0
        # 1D equivalent -- out[0] = (f[1] - f[0])
        out[tuple(slice1)] = f[tuple(slice2)] - f[tuple(slice3)]
        slice1[axis] = -1
        slice2[axis] = -1
        slice3[axis] = -2
        # 1D equivalent -- out[-1] = (f[-1] - f[-2])
        out[tuple(slice1)] = f[tuple(slice2)] - f[tuple(slice3)]

        # divide by step size
        outvals.append(out / dx[axis])
        # reset the slice object in this dimension to ":"
        slice1[axis] = slice(None)
        slice2[axis] = slice(None)
        slice3[axis] = slice(None)

    if N == 1:
        return outvals[0]
    else:
        return outvals


def streamline_gradients(streamline):
    """
    Calculate the gradients of the streamline along the spatial dimension

    Parameters
    ----------
    streamline : array-like of shape (n, 3)
        The 3d coordinates of a single streamline

    Returns
    -------
    Array of shape (3, n): Spatial gradients along the length of the
    streamline.

    """
    return np.array(gradient(np.asarray(streamline))[0])


def grad_tensor(grad, evals):
    """
    Calculate the 3 by 3 tensor for a given spatial gradient, given a canonical
    tensor shape (also as a 3 by 3), pointing at [1,0,0]

    Parameters
    ----------
    grad : 1d array of shape (3,)
        The spatial gradient (e.g between two nodes of a streamline).

    evals: 1d array of shape (3,)
        The eigenvalues of a canonical tensor to be used as a response
        function.

    """
    # This is the rotation matrix from [1, 0, 0] to this gradient of the sl:
    R = la.svd([grad], overwrite_a=True)[2]
    # This is the 3 by 3 tensor after rotation:
    T = np.linalg.multi_dot([R, np.diag(evals), R.T])
    return T


@warning_for_keywords()
def streamline_tensors(streamline, *, evals=(0.001, 0, 0)):
    """
    The tensors generated by this fiber.

    Parameters
    ----------
    streamline : array-like of shape (n, 3)
        The 3d coordinates of a single streamline

    evals : iterable with three entries, optional
        The estimated eigenvalues of a single fiber tensor.

    Returns
    -------
    An n_nodes by 3 by 3 array with the tensor for each node in the fiber.

    Notes
    -----
    Estimates of the radial/axial diffusivities may rely on
    empirical measurements (for example, the AD in the Corpus Callosum), or
    may be based on a biophysical model of some kind.
    """

    grad = streamline_gradients(streamline)

    # Preallocate:
    tensors = np.empty((grad.shape[0], 3, 3))

    for grad_idx, this_grad in enumerate(grad):
        tensors[grad_idx] = grad_tensor(this_grad, evals)
    return tensors


@warning_for_keywords()
def streamline_signal(streamline, gtab, *, evals=(0.001, 0, 0)):
    """
    The signal from a single streamline estimate along each of its nodes.

    Parameters
    ----------
    streamline : a single streamline
        Streamline data.

    gtab : GradientTable class instance
        Gradient table.

    evals : array-like of length 3, optional
        The eigenvalues of the canonical tensor used as an estimate of the
        signal generated by each node of the streamline.
    """
    # Gotta have those tensors:
    tensors = streamline_tensors(streamline, evals=evals)
    sig = np.empty((len(streamline), np.sum(~gtab.b0s_mask)))
    # Extract them once:
    bvecs = gtab.bvecs[~gtab.b0s_mask]
    bvals = gtab.bvals[~gtab.b0s_mask]
    for ii, tensor in enumerate(tensors):
        ADC = np.diag(np.linalg.multi_dot([bvecs, tensor, bvecs.T]))
        # Use the Stejskal-Tanner equation with the ADC as input, and S0 = 1:
        sig[ii] = np.exp(-bvals * ADC)
    return sig - np.mean(sig)


class LifeSignalMaker:
    """
    A class for generating signals from streamlines in an efficient and speedy
    manner.
    """

    @warning_for_keywords()
    def __init__(self, gtab, *, evals=(0.001, 0, 0), sphere=None):
        """
        Initialize a signal maker

        Parameters
        ----------
        gtab : GradientTable class instance
            The gradient table on which the signal is calculated.
        evals : array-like of 3 items, optional
            The eigenvalues of the canonical tensor to use in calculating the
            signal.
        sphere : `dipy.core.Sphere` class instance, optional
            The discrete sphere to use as an approximation for the continuous
            sphere on which the signal is represented. If integer - we will use
            an instance of one of the symmetric spheres cached in
            `dps.get_sphere`. If a 'dipy.core.Sphere' class instance is
            provided, we will use this object. Default: the :mod:`dipy.data`
            symmetric sphere with 724 vertices
        """
        self.sphere = sphere or dpd.default_sphere

        self.gtab = gtab
        self.evals = evals
        # Initialize an empty dict to fill with signals for each of the sphere
        # vertices:
        self.signal = np.empty((self.sphere.vertices.shape[0], np.sum(~gtab.b0s_mask)))
        # We'll need to keep track of what we've already calculated:
        self._calculated = []

    def calc_signal(self, xyz):
        idx = self.sphere.find_closest(xyz)
        if idx not in self._calculated:
            bvecs = self.gtab.bvecs[~self.gtab.b0s_mask]
            bvals = self.gtab.bvals[~self.gtab.b0s_mask]
            tensor = grad_tensor(self.sphere.vertices[idx], self.evals)
            ADC = np.diag(np.linalg.multi_dot([bvecs, tensor, bvecs.T]))
            sig = np.exp(-bvals * ADC)
            sig = sig - np.mean(sig)
            self.signal[idx] = sig
            self._calculated.append(idx)

        return self.signal[idx]

    def streamline_signal(self, streamline):
        """
        Approximate the signal for a given streamline
        """
        grad = streamline_gradients(streamline)
        sig_out = np.zeros((grad.shape[0], self.signal.shape[-1]))
        for ii, g in enumerate(grad):
            sig_out[ii] = self.calc_signal(g)
        return sig_out


@warning_for_keywords()
def voxel2streamline(streamline, affine, *, unique_idx=None):
    """
    Maps voxels to streamlines and streamlines to voxels, for setting up
    the LiFE equations matrix

    Parameters
    ----------
    streamline : list
        A collection of streamlines, each n by 3, with n being the number of
        nodes in the fiber.
    affine : array_like (4, 4)
        The mapping from voxel coordinates to streamline points.
        The voxel_to_rasmm matrix, typically from a NIFTI file.
    unique_idx : array, optional.
       The unique indices in the streamlines

    Returns
    -------
    v2f, v2fn : tuple of dicts

    The first dict in the tuple answers the question: Given a voxel (from
    the unique indices in this model), which fibers pass through it?

    The second answers the question: Given a streamline, for each voxel that
    this streamline passes through, which nodes of that streamline are in that
    voxel?
    """
    transformed_streamline = transform_streamlines(streamline, affine)

    if unique_idx is None:
        all_coords = np.concatenate(transformed_streamline)
        unique_idx = unique_rows(np.round(all_coords))

    return _voxel2streamline(transformed_streamline, unique_idx.astype(np.intp))


class FiberModel(ReconstModel):
    """
    A class for representing and solving predictive models based on
    tractography solutions.

    Notes
    -----
    This is an implementation of the LiFE model described in
    :footcite:p:`Pestilli2014`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, gtab):
        """
        Parameters
        ----------
        gtab : a GradientTable class instance

        """
        # Initialize the super-class:
        ReconstModel.__init__(self, gtab)

    @warning_for_keywords()
    def setup(self, streamline, affine, *, evals=(0.001, 0, 0), sphere=None):
        """
        Set up the necessary components for the LiFE model: the matrix of
        fiber-contributions to the DWI signal, and the coordinates of voxels
        for which the equations will be solved

        Parameters
        ----------
        streamline : list
            Streamlines, each is an array of shape (n, 3)
        affine : array_like (4, 4)
            The mapping from voxel coordinates to streamline points.
            The voxel_to_rasmm matrix, typically from a NIFTI file.
        evals : array-like (3 items, optional)
            The eigenvalues of the canonical tensor used as a response
            function. Default:[0.001, 0, 0].
        sphere: `dipy.core.Sphere` instance, optional
            Whether to approximate (and cache) the signal on a discrete
            sphere. This may confer a significant speed-up in setting up the
            problem, but is not as accurate. If `False`, we use the exact
            gradients along the streamlines to calculate the matrix, instead of
            an approximation. Defaults to use the 724-vertex symmetric sphere
            from :mod:`dipy.data`
        """
        if sphere is not False:
            SignalMaker = LifeSignalMaker(self.gtab, evals=evals, sphere=sphere)

        streamline = transform_streamlines(streamline, affine)
        # Assign some local variables, for shorthand:
        all_coords = np.concatenate(streamline)
        vox_coords = unique_rows(np.round(all_coords).astype(np.intp))
        del all_coords
        # We only consider the diffusion-weighted signals:
        n_bvecs = self.gtab.bvals[~self.gtab.b0s_mask].shape[0]
        v2f, v2fn = voxel2streamline(streamline, np.eye(4), unique_idx=vox_coords)
        # How many fibers in each voxel (this will determine how many
        # components are in the matrix):
        n_unique_f = len(np.hstack(list(v2f.values())))
        # Preallocate these, which will be used to generate the sparse
        # matrix:
        f_matrix_sig = np.zeros(n_unique_f * n_bvecs, dtype=float)
        f_matrix_row = np.zeros(n_unique_f * n_bvecs, dtype=np.intp)
        f_matrix_col = np.zeros(n_unique_f * n_bvecs, dtype=np.intp)

        fiber_signal = []
        for _, s in enumerate(streamline):
            if sphere is not False:
                fiber_signal.append(SignalMaker.streamline_signal(s))
            else:
                fiber_signal.append(streamline_signal(s, self.gtab, evals=evals))

        del streamline
        if sphere is not False:
            del SignalMaker

        keep_ct = 0
        range_bvecs = np.arange(n_bvecs).astype(int)
        # In each voxel:
        for v_idx in range(vox_coords.shape[0]):
            mat_row_idx = (range_bvecs + v_idx * n_bvecs).astype(np.intp)
            # For each fiber in that voxel:
            for f_idx in v2f[v_idx]:
                # For each fiber-voxel combination, store the row/column
                # indices in the pre-allocated linear arrays
                f_matrix_row[keep_ct : keep_ct + n_bvecs] = mat_row_idx
                f_matrix_col[keep_ct : keep_ct + n_bvecs] = f_idx

                vox_fiber_sig = np.zeros(n_bvecs)
                for node_idx in v2fn[f_idx][v_idx]:
                    # Sum the signal from each node of the fiber in that voxel:
                    vox_fiber_sig += fiber_signal[f_idx][node_idx]
                # And add the summed thing into the corresponding rows:
                f_matrix_sig[keep_ct : keep_ct + n_bvecs] += vox_fiber_sig
                keep_ct = keep_ct + n_bvecs

        del v2f, v2fn
        # Allocate the sparse matrix, using the more memory-efficient 'csr'
        # format:
        life_matrix = sps.csr_array((f_matrix_sig, [f_matrix_row, f_matrix_col]))

        return life_matrix, vox_coords

    def _signals(self, data, vox_coords):
        """
        Helper function to extract and separate all the signals we need to fit
        and evaluate a fit of this model

        Parameters
        ----------
        data : 4D array

        vox_coords: n by 3 array
            The coordinates into the data array of the fiber nodes.
        """
        # Fitting is done on the S0-normalized-and-demeaned diffusion-weighted
        # signal:
        idx_tuple = (vox_coords[:, 0], vox_coords[:, 1], vox_coords[:, 2])
        # We'll look at a 2D array, extracting the data from the voxels:
        vox_data = data[idx_tuple]
        weighted_signal = vox_data[:, ~self.gtab.b0s_mask]
        b0_signal = np.mean(vox_data[:, self.gtab.b0s_mask], -1)
        relative_signal = weighted_signal / b0_signal[:, None]

        # The mean of the relative signal across directions in each voxel:
        mean_sig = np.mean(relative_signal, -1)
        to_fit = (relative_signal - mean_sig[:, None]).ravel()
        to_fit[np.isnan(to_fit)] = 0
        return (to_fit, weighted_signal, b0_signal, relative_signal, mean_sig, vox_data)

    @warning_for_keywords()
    def fit(self, data, streamline, affine, *, evals=(0.001, 0, 0), sphere=None):
        """
        Fit the LiFE FiberModel for data and a set of streamlines associated
        with this data

        Parameters
        ----------
        data : 4D array
            Diffusion-weighted data
        streamline : list
            A bunch of streamlines
        affine : array_like (4, 4)
            The mapping from voxel coordinates to streamline points.
            The voxel_to_rasmm matrix, typically from a NIFTI file.
        evals : array-like, optional
            The eigenvalues of the tensor response function used in
            constructing the model signal. Default: [0.001, 0, 0]
        sphere: `dipy.core.Sphere` instance or False, optional
            Whether to approximate (and cache) the signal on a discrete
            sphere. This may confer a significant speed-up in setting up the
            problem, but is not as accurate. If `False`, we use the exact
            gradients along the streamlines to calculate the matrix, instead of
            an approximation.

        Returns
        -------
        FiberFit class instance
        """
        if affine is None:
            affine = np.eye(4)
        sl_len = np.array([len(s) for s in streamline])
        if np.any(sl_len < 2):
            raise ValueError(
                "Input contains streamlines with only one node."
                " The LiFE model cannot be fit with these streamlines included."
            )
        life_matrix, vox_coords = self.setup(
            streamline, affine, evals=evals, sphere=sphere
        )
        (to_fit, weighted_signal, b0_signal, relative_signal, mean_sig, vox_data) = (
            self._signals(data, vox_coords)
        )
        beta = opt.sparse_nnls(to_fit, life_matrix)
        return FiberFit(
            self,
            life_matrix,
            vox_coords,
            to_fit,
            beta,
            weighted_signal,
            b0_signal,
            relative_signal,
            mean_sig,
            vox_data,
            streamline,
            affine,
            evals,
        )


class FiberFit(ReconstFit):
    """
    A fit of the LiFE model to diffusion data
    """

    def __init__(
        self,
        fiber_model,
        life_matrix,
        vox_coords,
        to_fit,
        beta,
        weighted_signal,
        b0_signal,
        relative_signal,
        mean_sig,
        vox_data,
        streamline,
        affine,
        evals,
    ):
        """
        Parameters
        ----------
        fiber_model : A FiberModel class instance

        params : the parameters derived from a fit of the model to the data.

        """
        ReconstFit.__init__(self, fiber_model, vox_data)

        self.life_matrix = life_matrix
        self.vox_coords = vox_coords
        self.fit_data = to_fit
        self.beta = beta
        self.weighted_signal = weighted_signal
        self.b0_signal = b0_signal
        self.relative_signal = relative_signal
        self.mean_signal = mean_sig
        self.streamline = streamline
        self.affine = affine
        self.evals = evals

    @warning_for_keywords()
    def predict(self, *, gtab=None, S0=None):
        """
        Predict the signal

        Parameters
        ----------
        gtab : GradientTable
            Default: use self.gtab
        S0 : float or array
            The non-diffusion-weighted signal in the voxels for which a
            prediction is made. Default: use self.b0_signal

        Returns
        -------
        prediction : ndarray of shape (voxels, bvecs)
            An array with a prediction of the signal in each voxel/direction
        """
        # We generate the prediction and in each voxel, we add the
        # offset, according to the isotropic part of the signal, which was
        # removed prior to fitting:
        if gtab is None:
            _matrix = self.life_matrix
            gtab = self.model.gtab
        else:
            _model = FiberModel(gtab)
            _matrix, _ = _model.setup(self.streamline, self.affine, evals=self.evals)

        pred_weighted = np.reshape(
            opt.spdot(_matrix, self.beta),
            (self.vox_coords.shape[0], np.sum(~gtab.b0s_mask)),
        )

        pred = np.empty((self.vox_coords.shape[0], gtab.bvals.shape[0]))
        if S0 is None:
            S0 = self.b0_signal

        pred[..., gtab.b0s_mask] = S0[:, None]
        pred[..., ~gtab.b0s_mask] = (pred_weighted + self.mean_signal[:, None]) * S0[
            :, None
        ]

        return pred



import logging

import numpy as np

from dipy.testing.decorators import warning_for_keywords


def check_img_shapes(images):
    """Check if the images have same shapes. It also provides details about the
    volumes are same or not. If the shapes are not equal it will return False
    for both shape and volume.

    Parameters
    ----------
    images : list

    Returns
    -------
    tuple
        tuple[0] = True, if shapes are equal.
        tuple[1] = True, if volumes are equal.
    """

    if len(images) < 2:
        return (True, False)
    base_shape = images[0][0].shape[:3]

    volumed_data_shapes = []
    for img in images:
        data, _, _ = unpack_image(img)
        data_shape = data.shape[:3]
        if data.ndim == 4:
            volumed_data_shapes.append(data.shape[3])
        if base_shape != data_shape:
            return (False, False)

    return (True, len(set(volumed_data_shapes)) == 1)


def check_img_dtype(images):
    """Check supplied image dtype.

    If not supported numerical type, fallback to supported numerical types
    (either int32 or float 32). If non-numerical type, skip the data.

    Parameters
    ----------
    images : list
        Each image is tuple of (data, affine).

    Returns
    -------
    list
        Valid images from the provided images.
    """

    valid_images = []

    for idx, img in enumerate(images):
        data, affine, fname = unpack_image(img)
        if np.issubdtype(data.dtype, np.integer):
            if data.dtype != np.int32:
                msg = f"{data.dtype} is not supported, falling back to int32"
                logging.warning(msg)
                img = (data.astype(np.int32), affine, fname)
            valid_images.append(img)
        elif np.issubdtype(data.dtype, np.floating):
            if data.dtype != np.float64 and data.dtype != np.float32:
                msg = f"{data.dtype} is not supported, falling back to float32"
                logging.warning(msg)
                img = (data.astype(np.float32), affine, fname)
            valid_images.append(img)
        else:
            msg = f"skipping image {idx + 1}, passed image is not in numerical format"
            logging.warning(msg)

    return valid_images


def show_ellipsis(text, text_size, available_size):
    """Apply ellipsis to the text.

    Parameters
    ----------
    text : string
        Text required to be check for ellipsis.
    text_size : float
        Current size of the text in pixels.
    available_size : float
        Size available to fit the text. This will be used to truncate the text
        and show ellipsis.

    Returns
    -------
    string
        Text after processing for ellipsis.
    """
    if text_size > available_size:
        max_chars = int((available_size / text_size) * len(text))
        ellipsis_text = f"...{text[-(max_chars - 3) :]}"
        return ellipsis_text
    return text


def unpack_image(img):
    """Unpack provided image data.

    Standard way to handle different value images.

    Parameters
    ----------
    img : tuple
        An image can contain either (data, affine) or (data, affine, fname).

    Returns
    -------
    tuple
        If img with (data, affine) it will convert to (data, affine, None).
        Otherwise it will be passed as it is.
    """
    return _unpack_data(img)


def unpack_surface(surface):
    """Unpack surface data.

    Parameters
    ----------
    surface : tuple
        It either contains (vertices, faces) or (vertices, faces, fname).

    Returns
    -------
    tuple
        If surface with (vertices, faces) it will convert to (vertices, faces,
        None). Otherwise it will be passed as it is.
    """
    data = _unpack_data(surface)

    if data[0].shape[-1] != 3:
        raise ValueError(f"Vertices do not have correct shape: {data[0].shape}")
    if data[1].shape[-1] != 3:
        raise ValueError(f"Faces do not have correct shape: {data[1].shape}")
    return data


def _unpack_data(data, return_size=3):
    result = [*data]
    if len(data) < return_size:
        result += (return_size - len(data)) * [None]

    return result


@warning_for_keywords()
def check_peak_size(pams, *, ref_img_shape=None, sync_imgs=False):
    """Check shape of peaks.

    Parameters
    ----------
    pams : PeaksAndMetrics
        Peaks and metrics.
    ref_img_shape : tuple, optional
        3D shape of the image, by default None.
    sync_imgs : bool, optional
        True if the images are synchronized, by default False.

    Returns
    -------
    bool
        If the peaks are aligned with images and other peaks.
    """
    base_shape = pams[0].peak_dirs.shape[:3]

    for pam in pams:
        if pam.peak_dirs.shape[:3] != base_shape:
            return False

    if not ref_img_shape:
        return True

    return base_shape == ref_img_shape and sync_imgs



"""Module for different conversion functions."""


def expand_range(range_str):
    """Convert a string with ranges to a list of integers.

    Parameters
    ----------
    range_str : str
        string with ranges, e.g. '1,2,3-5,7'

    Returns
    -------
    list(int)

    """
    range_str = range_str.replace(" ", "").strip()
    range_str = range_str[:-1] if range_str.endswith(",") else range_str
    result = []
    for part in range_str.split(","):
        if not part.replace("-", "").isdigit():
            raise ValueError(f"Incorrect character found in input: {part}")
        if "-" in part:
            start, end = map(int, part.split("-"))
            result += list(range(start, end + 1))
        else:
            result.append(int(part))
    return result



"""
============================================
Tractography Clustering - Available Features
============================================

This page lists available features that can be used by the tractography
clustering framework. For every feature a brief description is provided
explaining: what it does, when it's useful and how to use it. If you are not
familiar with the tractography clustering framework, read the
:ref:`clustering-framework` first.

.. contents:: Available Features
    :local:
    :depth: 1


Let's import the necessary modules.
"""

import numpy as np

from dipy.segment.clustering import QuickBundles
from dipy.segment.featurespeed import (
    ArcLengthFeature,
    CenterOfMassFeature,
    IdentityFeature,
    MidpointFeature,
    ResampleFeature,
    VectorOfEndpointsFeature,
)
from dipy.segment.metric import (
    AveragePointwiseEuclideanMetric,
    CosineMetric,
    EuclideanMetric,
)
from dipy.tracking.streamline import set_number_of_points
from dipy.viz import actor, colormap as cmap, window

###############################################################################
# .. note::
#
#     All examples assume a function `get_streamlines` exists. We defined here
#     a simple function to do so. It imports the necessary modules and loads a
#     small streamline bundle.


def get_streamlines():
    from dipy.data import get_fnames
    from dipy.io.streamline import load_tractogram
    from dipy.tracking.streamline import Streamlines

    fname = get_fnames(name="fornix")
    fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

    streamlines = Streamlines(fornix)
    return streamlines


###############################################################################
# .. _clustering-examples-IdentityFeature:
#
# Identity Feature
# ================
# **What:** Instances of `IdentityFeature` simply return the streamlines
# unaltered.  In other words the features are the original data.
#
# **When:** The QuickBundles algorithm requires streamlines to have the same
# number of points. If this is the case for your streamlines, you can tell
# QuickBundles to not perform resampling (see following example). The
# clustering should be faster than using the default behaviour of QuickBundles
# since it will require less computation (i.e. no resampling). However, it
# highly depends on the number of points streamlines have. By default,
# QuickBundles resamples streamlines so that they have 12 points each
# :footcite:p:`Garyfallidis2012a`.
#
# *Unless stated otherwise, it is the default feature used by `Metric` objects
# in the clustering framework.*


# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

# Make sure our streamlines have the same number of points.
streamlines = set_number_of_points(streamlines, nb_points=12)

# Create an instance of `IdentityFeature` and tell metric to use it.
feature = IdentityFeature()
metric = AveragePointwiseEuclideanMetric(feature=feature)
qb = QuickBundles(threshold=10.0, metric=metric)
clusters = qb.cluster(streamlines)

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", list(map(len, clusters)))

###############################################################################
# .. _clustering-examples-ResampleFeature:
#
# Resample Feature
# ================
# **What:** Instances of `ResampleFeature` resample streamlines to a
# predetermined number of points. The resampling is done on the fly such that
# there are no permanent modifications made to your streamlines.
#
# **When:** The QuickBundles algorithm requires streamlines to have the same
# number of points. By default, QuickBundles uses `ResampleFeature` to resample
# streamlines so that they have 12 points each :footcite:p:`Garyfallidis2012a`.
# If you want to use a different number of points for the resampling, you should
# provide your own instance of `ResampleFeature` (see following example).
#
# **Note:** Resampling streamlines has an impact on clustering results both in
# terms of speed and quality. Setting the number of points too low will result
# in a loss of information about the shape of the streamlines. On the contrary,
# setting the number of points too high will slow down the clustering process.

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

# Streamlines will be resampled to 24 points on the fly.
feature = ResampleFeature(nb_points=24)
metric = AveragePointwiseEuclideanMetric(feature=feature)  # a.k.a. MDF
qb = QuickBundles(threshold=10.0, metric=metric)
clusters = qb.cluster(streamlines)

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", list(map(len, clusters)))

###############################################################################
# .. _clustering-examples-CenterOfMassFeature:
#
# Center of Mass Feature
# ======================
# **What:** Instances of `CenterOfMassFeature` compute the center of mass
# (also known as center of gravity) of a set of points. This is achieved by
# taking the mean of every coordinate independently (for more information see
# the `wiki page <https://en.wikipedia.org/wiki/Center_of_mass>`_).
#
# **When:** This feature can be useful when you *only* need information about
# the spatial position of a streamline.
#
# **Note:** The computed center is not guaranteed to be an existing point in
# the streamline.

# Enables/disables interactive visualization
interactive = False

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.


feature = CenterOfMassFeature()
metric = EuclideanMetric(feature)

qb = QuickBundles(threshold=5.0, metric=metric)
clusters = qb.cluster(streamlines)

# Extract feature of every streamline.
centers = np.asarray(list(map(feature.extract, streamlines)))

# Color each center of mass according to the cluster they belong to.
colormap = cmap.create_colormap(np.arange(len(clusters)))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

# Visualization
scene = window.Scene()
scene.clear()
scene.SetBackground(0, 0, 0)
scene.add(actor.streamtube(streamlines, colors=window.colors.white, opacity=0.05))
scene.add(actor.point(centers[:, 0, :], colormap_full, point_radius=0.2))
window.record(
    scene=scene, n_frames=1, out_path="center_of_mass_feature.png", size=(600, 600)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the center of mass of each streamline colored according to
# the QuickBundles results.
#
#
# .. _clustering-examples-MidpointFeature:
#
# Midpoint Feature
# ================
# **What:** Instances of `MidpointFeature` extract the middle point of a
# streamline. If there is an even number of points, the feature will then
# correspond to the point halfway between the two middle points.
#
# **When:** This feature can be useful when you *only* need information about
# the spatial position of a streamline. This can also be an alternative to the
# `CenterOfMassFeature` if the point extracted must be on the streamline.

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

feature = MidpointFeature()
metric = EuclideanMetric(feature)

qb = QuickBundles(threshold=5.0, metric=metric)
clusters = qb.cluster(streamlines)

# Extract feature of every streamline.
midpoints = np.asarray(list(map(feature.extract, streamlines)))

# Color each midpoint according to the cluster they belong to.
colormap = cmap.create_colormap(np.arange(len(clusters)))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

# Visualization
scene = window.Scene()
scene.clear()
scene.SetBackground(0, 0, 0)
scene.add(actor.point(midpoints[:, 0, :], colormap_full, point_radius=0.2))
scene.add(actor.streamtube(streamlines, colors=window.colors.white, opacity=0.05))
window.record(scene=scene, n_frames=1, out_path="midpoint_feature.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the middle point of each streamline colored according to the
# QuickBundles results.
#
#
# .. _clustering-examples-ArcLengthFeature:
#
# ArcLength Feature
# =================
# **What:** Instances of `ArcLengthFeature` compute the length of a streamline.
# More specifically, this feature corresponds to the sum of the lengths of
# every streamline's segments.
#
# **When:** This feature can be useful when you *only* need information about
# the length of a streamline.

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

feature = ArcLengthFeature()
metric = EuclideanMetric(feature)
qb = QuickBundles(threshold=2.0, metric=metric)
clusters = qb.cluster(streamlines)

# Color each streamline according to the cluster they belong to.
colormap = cmap.create_colormap(np.ravel(clusters.centroids))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

# Visualization
scene = window.Scene()
scene.clear()
scene.SetBackground(0, 0, 0)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="arclength_feature.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the streamlines colored according to their length.
#
#
# .. _clustering-examples-VectorOfEndpointsFeature:
#
# Vector Between Endpoints Feature
# ================================
# **What:** Instances of `VectorOfEndpointsFeature` extract the vector going
# from one extremity of the streamline to the other. In other words, this
# feature represents the vector beginning at the first point and ending at the
# last point of the streamlines.
#
# **When:** This feature can be useful when you *only* need information about
# the orientation of a streamline.
#
# **Note:** Since streamlines endpoints are ambiguous (e.g. the first point
# could be either the beginning or the end of the streamline), one must be
# careful when using this feature.

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

feature = VectorOfEndpointsFeature()
metric = CosineMetric(feature)
qb = QuickBundles(threshold=0.1, metric=metric)
clusters = qb.cluster(streamlines)

# Color each streamline according to the cluster they belong to.
colormap = cmap.create_colormap(np.arange(len(clusters)))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

# Visualization
scene = window.Scene()
scene.clear()
scene.SetBackground(0, 0, 0)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="vector_of_endpoints_feature.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the streamlines colored according to their orientation.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
Title : Denoising diffusion weighted imaging data using CNN
===========================================================

Obtaining tissue microstructure measurements from diffusion weighted imaging
(DWI) with multiple, high b-values is crucial. However, the high noise levels
present in these images can adversely affect the accuracy of the
microstructural measurements. In this context, we suggest a straightforward
denoising technique :footcite:p:`Cheng2022` that can be applied to any DWI
dataset as long as a low-noise, single-subject dataset is obtained using the
same DWI sequence.

We created a simple 1D-CNN model with five layers, based on the 1D CNN for
denoising speech. The model consists of two convolutional layers followed by
max-pooling layers, and a dense layer. The first convolutional layer has
16 one-dimensional filters of size 16, and the second layer has 32 filters of
size 8. ReLu activation function is applied to both convolutional layers.
The max-pooling layer has a kernel size of 2 and a stride of 2.
The dense layer maps the features extracted from the noisy image to the
low-noise reference image.

Reference
---------

.. footbibliography::
"""

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
if have_tf:
    from tensorflow.keras.initializers import Orthogonal
    from tensorflow.keras.layers import Activation, Conv1D, Input
    from tensorflow.keras.models import Model

sklearn, have_sklearn, _ = optional_package("sklearn.model_selection")


class Cnn1DDenoiser:
    @warning_for_keywords()
    def __init__(
        self,
        sig_length,
        *,
        optimizer="adam",
        loss="mean_squared_error",
        metrics=("accuracy",),
        loss_weights=None,
    ):
        """Initialize the CNN 1D denoiser with the given parameters.

        Parameters
        ----------
        sig_length : int
            Length of the DWI signal.
        optimizer : str, optional
            Name of the optimization algorithm to use. Options: 'adam', 'sgd',
            'rmsprop', 'adagrad', 'adadelta'.
        loss : str, optional
            Name of the loss function to use. Available options are
            'mean_squared_error', 'mean_absolute_error',
            'mean_absolute_percentage_error', 'mean_squared_logarithmic_error',
            'squared_hinge', 'hinge', 'categorical_hinge', 'logcosh',
            'categorical_crossentropy', 'sparse_categorical_crossentropy',
            'binary_crossentropy', 'kullback_leibler_divergence', 'poisson',
            'cosine_similarity'.
            Suggested to go with 'mean_squared_error'.
        metrics : tuple of str or function, optional
            List of metrics to be evaluated by the model during training and
            testing. Available options are 'accuracy', 'binary_accuracy',
            'categorical_accuracy', 'top_k_categorical_accuracy',
            'sparse_categorical_accuracy', 'sparse_top_k_categorical_accuracy',
            and any custom function.
        loss_weights : float or dict, optional
            Scalar coefficients to weight the loss contributions of different
            model outputs. Can be a single float value or a dictionary mapping
            output names to scalar coefficients.

        """
        if not have_tf:
            raise ImportError(
                "TensorFlow is not available. Please install TensorFlow 2+."
            )
        if not have_sklearn:
            raise ImportError(
                "scikit-learn is not available. Please install scikit-learn."
            )
        input_layer = Input(shape=(sig_length, 1))
        x = Conv1D(
            filters=16,
            kernel_size=16,
            kernel_initializer=Orthogonal(),
            padding="same",
            name="Conv1",
        )(input_layer)
        x = Activation("relu", name="ReLU1")(x)
        max_pool_1d = tf.keras.layers.MaxPooling1D(
            pool_size=2, strides=2, padding="valid"
        )
        pool1 = max_pool_1d(x)
        x = Conv1D(filters=32, kernel_size=8, padding="same", name="Conv2")(pool1)
        x = Activation("relu", name="ReLU2")(x)
        pool2 = max_pool_1d(x)
        pool2_flat = tf.keras.layers.Flatten()(pool2)
        logits = tf.keras.layers.Dense(units=sig_length, activation="relu")(pool2_flat)
        model = Model(inputs=input_layer, outputs=logits)
        model.compile(
            optimizer=optimizer, loss=loss, metrics=metrics, loss_weights=loss_weights
        )
        self.model = model

    @warning_for_keywords()
    def compile(self, *, optimizer="adam", loss=None, metrics=None, loss_weights=None):
        """Configure the model for training.

        Parameters
        ----------
        optimizer : str or optimizer object, optional
            Name of optimizer or optimizer object.
        loss : str or objective function, optional
            Name of objective function or objective function itself.
            If 'None', the model will be compiled without any loss function
            and can only be used to predict output.
        metrics : list of metrics, optional
            List of metrics to be evaluated by the model during training
            and testing.
        loss_weights : list or dict, optional
            Optional list or dictionary specifying scalar coefficients(floats)
            to weight the loss contributions of different model outputs.
            The loss value that will be minimized by the model will then be
            the weighted sum of all individual losses. If a list, it is
            expected to have a 1:1 mapping to the model's outputs. If a dict,
            it is expected to map output names (strings) to scalar
            coefficients.

        """
        self.model.compile(
            optimizer=optimizer, loss=loss, metrics=metrics, loss_weights=loss_weights
        )

    def summary(self):
        """Get the summary of the model.

        The summary is textual and includes information about:
        The layers and their order in the model.
        The output shape of each layer.

        Returns
        -------
        summary : NoneType
            the summary of the model
        """
        return self.model.summary()

    @warning_for_keywords()
    def train_test_split(
        self,
        x,
        y,
        *,
        test_size=None,
        train_size=None,
        random_state=None,
        shuffle=True,
        stratify=None,
    ):
        """Split the input data into random train and test subsets.

        Parameters
        ----------
        x: numpy array
           input data.
        y: numpy array
           target data.
        test_size: float or int, optional
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the test split.
            If int, represents the absolute number of test samples.
            If None, the value is set to the complement of the train size.
            If train_size is also None, it will be set to 0.25.
        train_size: float or int, optional
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the train split.
            If int, represents the absolute number of train samples.
            If None, the value is automatically set to the complement of the
            test size.
        random_state: int, np.random.Generator instance or None, optional
            Controls the shuffling applied to the data before applying
            the split. Pass an int for reproducible output across multiple
            function calls. See Glossary.
        shuffle: bool, optional
            Whether or not to shuffle the data before splitting.
            If shuffle=False then stratify must be None.
        stratify: array-like, optional
            If not None, data is split in a stratified fashion,
            using this as the class labels. Read more in the User Guide.

        Returns
        -------
        Tuple of four numpy arrays: x_train, x_test, y_train, y_test.
        """
        sz = x.shape
        if len(sz) == 4:
            x = np.reshape(x, (sz[0] * sz[1] * sz[2], sz[3]))
        sz = y.shape
        if len(sz) == 4:
            y = np.reshape(y, (sz[0] * sz[1] * sz[2], sz[3]))
        return sklearn.train_test_split(
            x,
            y,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify,
        )

    @warning_for_keywords()
    def fit(
        self,
        x,
        y,
        *,
        batch_size=None,
        epochs=1,
        verbose=1,
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_batch_size=None,
        validation_freq=1,
    ):
        """Train the model on train dataset.

        The fit method will train the model for a fixed number of epochs
        (iterations) on a dataset. If given data is  4D it will convert
        it into 1D.

        Parameters
        ----------
        x : ndarray
            The input data, as an ndarray.
        y : ndarray
            The target data, as an ndarray.
        batch_size : int or None, optional
            Number of samples per batch of computation.
        epochs : int, optional
            The number of epochs.
        verbose : 'auto', 0, 1, or 2, optional
            Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line
            per epoch.
        callbacks : list of keras.callbacks.Callback instances, optional
            List of callbacks to apply during training.
        validation_split : float between 0 and 1, optional
            Fraction of the training data to be used as validation data.
        validation_data : tuple (x_val, y_val) or None, optional
            Data on which to evaluate the loss and any model metrics at
            the end of each epoch.
        shuffle : boolean, optional
            This argument is ignored when x is a generator or an object of
            tf.data.Dataset.
        initial_epoch : int, optional
            Epoch at which to start training.
        steps_per_epoch : int or None, optional
            Total number of steps (batches of samples) before declaring one
            epoch finished and starting the next epoch.
        validation_batch_size : int or None, optional
            Number of samples per validation batch.
        validation_steps : int or None, optional
            Only relevant if validation_data is provided and is a
            tf.data dataset.
        validation_freq : int or list/tuple/set, optional
            Only relevant if validation data is provided. If an integer,
            specifies how many training epochs to run before a new validation
            run is performed. If a list, tuple, or set, specifies the epochs
            on which to run validation.
        max_queue_size : int, optional
            Used for generator or keras.utils.Sequence input only.
        workers : integer, optional
            Used for generator or keras.utils.Sequence input only.
        use_multiprocessing : boolean, optional
            Used for generator or keras.utils.Sequence input only.

        Returns
        -------
        hist : object
            A History object. Its History.history attribute is a record of
            training loss values and metrics values at successive epochs.

        """
        sz = x.shape
        if len(sz) == 4:
            x = np.reshape(x, (sz[0] * sz[1] * sz[2], sz[3]))
        sz = y.shape
        if len(sz) == 4:
            y = np.reshape(y, (sz[0] * sz[1] * sz[2], sz[3]))
        return self.model.fit(
            x=x,
            y=y,
            batch_size=batch_size,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            validation_split=validation_split,
            validation_data=validation_data,
            shuffle=shuffle,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            validation_batch_size=validation_batch_size,
            validation_freq=validation_freq,
        )

    @warning_for_keywords()
    def evaluate(
        self,
        x,
        y,
        *,
        batch_size=None,
        verbose=1,
        steps=None,
        callbacks=None,
        return_dict=False,
    ):
        """Evaluate the model on a test dataset.

        Parameters
        ----------
        x : ndarray
            Test dataset (high-noise data). If 4D, it will be converted to 1D.
        y : ndarray
            Labels of the test dataset (low-noise data). If 4D, it will be
            converted to 1D.
        batch_size : int, optional
            Number of samples per gradient update.
        verbose : int, optional
            Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line
            per epoch.
        steps : int, optional
            Total number of steps (batches of samples) before declaring the
            evaluation round finished.
        callbacks : list, optional
            List of callbacks to apply during evaluation.
        max_queue_size : int, optional
            Maximum size for the generator queue.
        workers : int, optional
            Maximum number of processes to spin up when using process-based
            threading.
        use_multiprocessing : bool, optional
            If `True`, use process-based threading.
        return_dict : bool, optional
            If `True`, loss and metric results are returned as a dictionary.

        Returns
        -------
        List or dict
            If `return_dict` is `False`, returns a list of [loss, metrics]
            values on the test dataset. If `return_dict` is `True`, returns
            a dictionary of metric names and their corresponding values.

        """
        sz = x.shape
        if len(sz) == 4:
            x = np.reshape(x, (sz[0] * sz[1] * sz[2], sz[3]))
        sz = y.shape
        if len(sz) == 4:
            y = np.reshape(y, (sz[0] * sz[1] * sz[2], sz[3]))
        return self.model.evaluate(
            x=x,
            y=y,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=callbacks,
            return_dict=return_dict,
        )

    @warning_for_keywords()
    def predict(self, x, *, batch_size=None, verbose=0, steps=None, callbacks=None):
        """Generate predictions for input samples.

        Parameters
        ----------
        x : ndarray
            Input samples.
        batch_size : int, optional
            Number of samples per batch.
        verbose : int, optional
            Verbosity mode.
        steps : int, optional
            Total number of steps (batches of samples) before declaring the
            prediction round finished.
        callbacks : list, optional
            List of Keras callbacks to apply during prediction.
        max_queue_size : int, optional
            Maximum size for the generator queue.
        workers : int, optional
            Maximum number of processes to spin up when using process-based
            threading.
        use_multiprocessing : bool, optional
            If `True`, use process-based threading. If `False`, use
            thread-based threading.

        Returns
        -------
        ndarray
            Numpy array of predictions.

        """
        sz = x.shape
        x = np.reshape(x, (sz[0] * sz[1] * sz[2], sz[3]))
        predicted_output = self.model.predict(
            x=x,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=callbacks,
        )
        predicted_output = np.float32(
            np.reshape(predicted_output, (sz[0], sz[1], sz[2], sz[3]))
        )
        return predicted_output

    @warning_for_keywords()
    def save_weights(self, filepath, *, overwrite=True):
        """Save the weights of the model to HDF5 file format.

        Parameters
        ----------
        filepath : str
            The path where the weights should be saved.
        overwrite : bool,optional
            If `True`, overwrites the file if it already exists. If `False`,
            raises an error if the file already exists.

        """
        self.model.save_weights(
            filepath=filepath, overwrite=overwrite, save_format=None
        )

    def load_weights(self, filepath):
        """Load the model weights from the specified file path.

        Parameters
        ----------
        filepath : str
            The file path from which to load the weights.

        """
        self.model.load_weights(filepath)



import numpy.testing as npt
import pytest

from dipy.align.imwarp import SymmetricDiffeomorphicRegistration
from dipy.align.metrics import SSDMetric
from dipy.testing.decorators import set_random_number_generator

# Conditional import machinery for matplotlib
from dipy.utils.optpkg import optional_package
from dipy.viz import regtools

_, have_matplotlib, _ = optional_package("matplotlib")


@pytest.mark.skipif(not have_matplotlib, reason="Requires Matplotlib")
@set_random_number_generator()
def test_plot_2d_diffeomorphic_map(rng):
    # Test the regtools plotting interface (lightly).
    mv_shape = (11, 12)
    moving = rng.random(mv_shape)
    st_shape = (13, 14)
    static = rng.random(st_shape)
    dim = static.ndim
    metric = SSDMetric(dim)
    level_iters = [200, 100, 50, 25]
    sdr = SymmetricDiffeomorphicRegistration(
        metric=metric, level_iters=level_iters, inv_iter=50
    )
    mapping = sdr.optimize(static, moving)
    # Smoke testing of plots
    ff = regtools.plot_2d_diffeomorphic_map(mapping, delta=10)
    # Default shape is static shape, moving shape
    npt.assert_equal(ff[0].shape, st_shape)
    npt.assert_equal(ff[1].shape, mv_shape)
    # Can specify shape
    ff = regtools.plot_2d_diffeomorphic_map(
        mapping, delta=10, direct_grid_shape=(7, 8), inverse_grid_shape=(9, 10)
    )
    npt.assert_equal(ff[0].shape, (7, 8))
    npt.assert_equal(ff[1].shape, (9, 10))



from glob import glob
import inspect
import itertools
import os

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.workflows.base import get_args_default


def common_start(sa, sb):
    """Return the longest common substring from the beginning of sa and sb."""

    def _iter():
        for a, b in zip(sa, sb):
            if a == b:
                yield a
            else:
                return

    return "".join(_iter())


def slash_to_under(dir_str):
    return "".join(dir_str.replace("/", "_"))


@warning_for_keywords()
def connect_output_paths(
    inputs, out_dir, out_files, *, output_strategy="absolute", mix_names=True
):
    """Generate a list of output files paths based on input files and
    output strategies.

    Parameters
    ----------
    inputs : array
        List of input paths.
    out_dir : string
        The output directory.
    out_files : array
        List of output files.
    output_strategy : string, optional
        Which strategy to use to generate the output paths.
            'append': Add out_dir to the path of the input.
            'prepend': Add the input path directory tree to out_dir.
            'absolute': Put directly in out_dir.
    mix_names : bool, optional
        Whether or not prepend a string composed of a mix of the input
        names to the final output name.

    Returns
    -------
        A list of output file paths.

    """
    outputs = []
    if isinstance(inputs, str):
        inputs = [inputs]
    if isinstance(out_files, str):
        out_files = [out_files]

    sizes_of_inputs = [len(inp) for inp in inputs]

    max_size = np.max(sizes_of_inputs)
    min_size = np.min(sizes_of_inputs)
    if min_size > 1 and min_size != max_size:
        raise ImportError("Size of input issue")

    elif min_size == 1:
        for i, sz in enumerate(sizes_of_inputs):
            if sz == min_size:
                inputs[i] = max_size * inputs[i]

    if mix_names:
        mixing_prefixes = concatenate_inputs(inputs)
    else:
        mixing_prefixes = [""] * len(inputs[0])

    for mix_pref, inp in zip(mixing_prefixes, inputs[0]):
        inp_dirname = os.path.dirname(inp)
        if output_strategy == "prepend":
            if os.path.isabs(out_dir):
                dname = out_dir + inp_dirname
            if not os.path.isabs(out_dir):
                dname = os.path.join(os.getcwd(), out_dir + inp_dirname)

        elif output_strategy == "append":
            dname = os.path.join(inp_dirname, out_dir)

        else:
            dname = out_dir

        updated_out_files = []
        for out_file in out_files:
            updated_out_files.append(os.path.join(dname, mix_pref + out_file))

        outputs.append(updated_out_files)

    return inputs, outputs


def concatenate_inputs(multi_inputs):
    """Concatenate list of inputs."""
    mixing_names = []
    for inps in zip(*multi_inputs):
        mixing_name = ""
        for inp in inps:
            mixing_name += basename_without_extension(inp) + "_"

        mixing_names.append(mixing_name + "_")
    return mixing_names


def basename_without_extension(fname):
    base = os.path.basename(fname)
    result = base.split(".")[0]
    if result[-4:] == ".nii":
        result = result.split(".")[0]

    return result


@warning_for_keywords()
def io_iterator(
    inputs,
    out_dir,
    fnames,
    *,
    output_strategy="absolute",
    mix_names=False,
    out_keys=None,
):
    """Create an IOIterator from the parameters.

    Parameters
    ----------
    inputs : array
        List of input files.
    out_dir : string
        Output directory.
    fnames : array
        File names of all outputs to be created.
    output_strategy : string, optional
        Controls the behavior of the IOIterator for output paths.
    mix_names : bool, optional
        Whether or not to append a mix of input names at the beginning.
    out_keys : list, optional
        Output parameter names.

    Returns
    -------
        Properly instantiated IOIterator object.

    """
    io_it = IOIterator(output_strategy=output_strategy, mix_names=mix_names)
    io_it.set_inputs(*inputs)
    io_it.set_out_dir(out_dir)
    io_it.set_out_fnames(*fnames)
    io_it.create_outputs()
    if out_keys:
        io_it.set_output_keys(*out_keys)

    return io_it


@warning_for_keywords()
def _io_iterator(frame, fnc, *, output_strategy="absolute", mix_names=False):
    """Create an IOIterator using introspection.

    Parameters
    ----------
    frame : frameobject
        Contains the info about the current local variables values.
    fnc : function
        The function to inspect
    output_strategy : string, optional
        Controls the behavior of the IOIterator for output paths.
    mix_names : bool, optional
        Whether or not to append a mix of input names at the beginning.

    Returns
    -------
        Properly instantiated IOIterator object.

    """

    # Create a new object that does not contain the ``self`` dict item
    def _selfless_dict(_values):
        return {key: val for key, val in _values.items() if key != "self"}

    args, _, _, values = inspect.getargvalues(frame)
    args.remove("self")
    # Create a new object that does not contain the ``self`` dict item from the
    # provided copy of the local symbol table returned by ``getargvalues``.
    # Avoids attempting to remove it from the object returned by
    # ``getargvalues``.
    values = _selfless_dict(values)

    spargs, defaults = get_args_default(fnc)

    len_args = len(spargs)
    len_defaults = len(defaults)
    split_at = len_args - len_defaults

    inputs = []
    outputs = []
    out_dir = ""

    # inputs
    for arv in args[:split_at]:
        inputs.append(values[arv])

    # defaults
    out_keys = []
    for arv in args[split_at:]:
        if arv == "out_dir":
            out_dir = values[arv]
        elif "out_" in arv:
            out_keys.append(arv)
            outputs.append(values[arv])

    return io_iterator(
        inputs,
        out_dir,
        outputs,
        output_strategy=output_strategy,
        mix_names=mix_names,
        out_keys=out_keys,
    )


class IOIterator:
    """Create output filenames that work nicely with multiple input files from
    multiple directories (processing multiple subjects with one command)

    Use information from input files, out_dir and out_fnames to generate
    correct outputs which can come from long lists of multiple or single
    inputs.
    """

    @warning_for_keywords()
    def __init__(self, *, output_strategy="absolute", mix_names=False):
        self.output_strategy = output_strategy
        self.mix_names = mix_names
        self.inputs = []
        self.out_keys = None

    def set_inputs(self, *args):
        self.file_existence_check(args)
        self.input_args = list(args)
        for inp in self.input_args:
            if isinstance(inp, str):
                self.inputs.append(sorted(glob(inp)))
            if isinstance(inp, list) and all(isinstance(s, str) for s in inp):
                nested = [sorted(glob(i)) for i in inp if isinstance(i, str)]
                self.inputs.append(list(itertools.chain.from_iterable(nested)))

    def set_out_dir(self, out_dir):
        self.out_dir = out_dir

    def set_out_fnames(self, *args):
        self.out_fnames = list(args)

    def set_output_keys(self, *args):
        self.out_keys = list(args)

    def create_outputs(self):
        if len(self.inputs) >= 1:
            self.updated_inputs, self.outputs = connect_output_paths(
                self.inputs,
                self.out_dir,
                self.out_fnames,
                output_strategy=self.output_strategy,
                mix_names=self.mix_names,
            )

            self.create_directories()

        else:
            raise ImportError("No inputs")

    def create_directories(self):
        for outputs in self.outputs:
            for output in outputs:
                directory = os.path.dirname(output)
                if not (directory == "" or os.path.exists(directory)):
                    os.makedirs(directory)

    def __iter__(self):
        ins = np.array(self.inputs).T
        out = np.array(self.outputs)
        IO = np.concatenate([ins, out], axis=1)
        for i_o in IO:
            if len(i_o) == 1:
                yield str(*i_o)
            else:
                yield i_o

    def file_existence_check(self, args):
        input_args = []
        for fname in args:
            if isinstance(fname, str):
                input_args.append(fname)
            # unpack variable string
            if isinstance(fname, list) and all(isinstance(s, str) for s in fname):
                input_args += fname
        for path in input_args:
            if len(glob(path)) == 0:
                raise OSError(f"File not found: {path}")



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False

cimport numpy as cnp

from libc.stdlib cimport rand, srand, RAND_MAX
from libc.math cimport sqrt


cdef void take(
        double* odf,
        cnp.npy_intp* indices,
        cnp.npy_intp n_indices,
        double* values_out) noexcept nogil


# Replaces a numpy.searchsorted(arr, number, 'right')
cdef int where_to_insert(
        cnp.float_t* arr,
        cnp.float_t number,
        int size) noexcept nogil

cdef void cumsum(
        cnp.float_t* arr_in,
        cnp.float_t* arr_out,
        int N) noexcept nogil

cdef void copy_point(
        double * a,
        double * b) noexcept nogil

cdef void scalar_muliplication_point(
        double * a,
        double scalar) noexcept nogil

cdef double norm(
        double * v) noexcept nogil

cdef double dot(
        double * v1,
        double * v2) noexcept nogil

cdef void normalize(
        double * v) noexcept nogil

cdef void cross(
        double * out,
        double * v1,
        double * v2) noexcept nogil

cdef void random_vector(
        double * out,
        RNGState* rng=*) noexcept nogil

cdef void random_perpendicular_vector(
        double * out,
        double * v,
        RNGState* rng=*) noexcept nogil

cdef (double, double) random_point_within_circle(
        double r, RNGState* rng=*) noexcept nogil

cpdef double random() noexcept nogil

cpdef void seed(cnp.npy_uint32 s) noexcept nogil

cdef struct RNGState:
    cnp.npy_uint64 state
    cnp.npy_uint64 inc

cdef void seed_rng(RNGState* rng_state, cnp.npy_uint64 seed) noexcept nogil

cdef cnp.npy_uint32 next_rng(RNGState* rng_state) noexcept nogil

cdef double random_float(RNGState* rng_state) noexcept nogil



"""pytest initialization."""

import importlib
import os
import re
import warnings

import numpy as np
import pytest

""" Set numpy print options to "legacy" for new versions of numpy
 If imported into a file, pytest will run this before any doctests.

References
----------
https://github.com/numpy/numpy/commit/710e0327687b9f7653e5ac02d222ba62c657a718
https://github.com/numpy/numpy/commit/734b907fc2f7af6e40ec989ca49ee6d87e21c495
https://github.com/nipy/nibabel/pull/556
"""
np.set_printoptions(legacy="1.13")

warnings.simplefilter(action="default", category=FutureWarning)
# List of files that pytest should ignore
collect_ignore = ["testing/decorators.py", "bench*.py", "**/benchmarks/*"]


def pytest_addoption(parser):
    parser.addoption(
        "--warnings-as-errors",
        action="store_true",
        help="Make all uncaught warnings into errors.",
    )


@pytest.hookimpl(trylast=True)
def pytest_sessionfinish(session, exitstatus):
    have_werrors = os.getenv("DIPY_WERRORS", False)
    have_werrors = (
        session.config.getoption("--warnings-as-errors", False) or have_werrors
    )
    if have_werrors:
        # Check if there were any warnings during the test session
        reporter = session.config.pluginmanager.get_plugin("terminalreporter")
        if reporter.stats.get("warnings", None):
            session.exitstatus = 2


@pytest.hookimpl
def pytest_terminal_summary(terminalreporter, exitstatus, config):
    have_werrors = os.getenv("DIPY_WERRORS", False)
    have_werrors = config.getoption("--warnings-as-errors", False) or have_werrors
    have_warnings = terminalreporter.stats.get("warnings", None)
    if have_warnings and have_werrors:
        terminalreporter.ensure_newline()
        terminalreporter.section("Werrors", sep="=", red=True, bold=True)
        terminalreporter.line(
            "Warnings as errors: Activated. \n"
            f"{len(have_warnings)} warnings were raised and "
            "treated as errors. \n"
        )


def pytest_collect_file(parent, file_path):
    if file_path.suffix in [".pyx", ".so"] and file_path.name.startswith("test"):
        return PyxFile.from_parent(parent, path=file_path)


class PyxFile(pytest.File):
    def collect(self):
        try:
            match = re.search(r"(dipy/[^\/]+/tests/test_\w+)", str(self.path))
            mod_name = match.group(1) if match else None
            if mod_name is None:
                raise PyxException("Could not find test module for " f"{self.path}.")
            mod_name = mod_name.replace("/", ".")
            mod = importlib.import_module(mod_name)
            for name in dir(mod):
                item = getattr(mod, name)
                if callable(item) and name.startswith("test_"):
                    yield PyxItem.from_parent(self, name=name, test_func=item, mod=mod)
        except ImportError as e:
            msg = (
                f"Import failed for {self.path}. Make sure you cython file "
                "has been compiled."
            )
            raise PyxException(msg, self.path, 0) from e


class PyxItem(pytest.Item):
    def __init__(self, *, test_func, mod, **kwargs):
        super().__init__(**kwargs)
        self.mod = mod
        self.test_func = test_func

    def runtest(self):
        """Called to execute the test item."""
        self.test_func()

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        return excinfo.value.args[0]

    def reportinfo(self):
        return self.path, 0, f"test: {self.name}"


class PyxException(Exception):
    """Custom exception for error reporting."""



#!/usr/bin/python
"""Class and helper functions for fitting the DeepN4 model."""

import logging

import numpy as np
from scipy.ndimage import gaussian_filter

from dipy.data import get_fnames
from dipy.nn.utils import normalize, recover_img, set_logger_level, transform_img
from dipy.testing.decorators import doctest_skip_parser, warning_for_keywords
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
if have_tf:
    from tensorflow.keras.layers import (
        Concatenate,
        Conv3D,
        Conv3DTranspose,
        GroupNormalization,
        Layer,
        LeakyReLU,
        MaxPool3D,
    )
    from tensorflow.keras.models import Model
else:

    class Model:
        pass

    class Layer:
        pass

    logging.warning(
        "This model requires Tensorflow.\
                    Please install these packages using \
                    pip. If using mac, please refer to this \
                    link for installation. \
                    https://github.com/apple/tensorflow_macos"
    )


logging.basicConfig()
logger = logging.getLogger("deepn4")


class EncoderBlock(Layer):
    def __init__(self, out_channels, kernel_size, strides, padding):
        super(EncoderBlock, self).__init__()
        self.conv3d = Conv3D(
            out_channels, kernel_size, strides=strides, padding=padding, use_bias=False
        )
        self.instnorm = GroupNormalization(
            groups=-1, axis=-1, epsilon=1e-05, center=False, scale=False
        )
        self.activation = LeakyReLU(0.01)

    def call(self, input):
        x = self.conv3d(input)
        x = self.instnorm(x)
        x = self.activation(x)

        return x


class DecoderBlock(Layer):
    def __init__(self, out_channels, kernel_size, strides, padding):
        super(DecoderBlock, self).__init__()
        self.conv3d = Conv3DTranspose(
            out_channels, kernel_size, strides=strides, padding=padding, use_bias=False
        )
        self.instnorm = GroupNormalization(
            groups=-1, axis=-1, epsilon=1e-05, center=False, scale=False
        )
        self.activation = LeakyReLU(0.01)

    def call(self, input):
        x = self.conv3d(input)
        x = self.instnorm(x)
        x = self.activation(x)

        return x


def UNet3D(input_shape):
    inputs = tf.keras.Input(input_shape)
    # Encode
    x = EncoderBlock(32, kernel_size=3, strides=1, padding="same")(inputs)
    syn0 = EncoderBlock(64, kernel_size=3, strides=1, padding="same")(x)

    x = MaxPool3D()(syn0)
    x = EncoderBlock(64, kernel_size=3, strides=1, padding="same")(x)
    syn1 = EncoderBlock(128, kernel_size=3, strides=1, padding="same")(x)

    x = MaxPool3D()(syn1)
    x = EncoderBlock(128, kernel_size=3, strides=1, padding="same")(x)
    syn2 = EncoderBlock(256, kernel_size=3, strides=1, padding="same")(x)

    x = MaxPool3D()(syn2)
    x = EncoderBlock(256, kernel_size=3, strides=1, padding="same")(x)
    x = EncoderBlock(512, kernel_size=3, strides=1, padding="same")(x)

    # Last layer without relu
    x = Conv3D(512, kernel_size=1, strides=1, padding="same")(x)

    x = DecoderBlock(512, kernel_size=2, strides=2, padding="valid")(x)

    x = Concatenate()([x, syn2])

    x = DecoderBlock(256, kernel_size=3, strides=1, padding="same")(x)
    x = DecoderBlock(256, kernel_size=3, strides=1, padding="same")(x)
    x = DecoderBlock(256, kernel_size=2, strides=2, padding="valid")(x)

    x = Concatenate()([x, syn1])

    x = DecoderBlock(128, kernel_size=3, strides=1, padding="same")(x)
    x = DecoderBlock(128, kernel_size=3, strides=1, padding="same")(x)
    x = DecoderBlock(128, kernel_size=2, strides=2, padding="valid")(x)

    x = Concatenate()([x, syn0])

    x = DecoderBlock(64, kernel_size=3, strides=1, padding="same")(x)
    x = DecoderBlock(64, kernel_size=3, strides=1, padding="same")(x)

    x = DecoderBlock(1, kernel_size=1, strides=1, padding="valid")(x)

    # Last layer without relu
    out = Conv3DTranspose(1, kernel_size=1, strides=1, padding="valid")(x)

    return Model(inputs, out)


class DeepN4:
    """This class is intended for the DeepN4 model.

    The DeepN4 model :footcite:p:`Kanakaraj2024` predicts the bias field for
    magnetic field inhomogeneity correction on T1-weighted images.

    References
    ----------
    .. footbibliography::
    """

    @warning_for_keywords()
    @doctest_skip_parser
    def __init__(self, *, verbose=False):
        """Model initialization

        To obtain the pre-trained model, use fetch_default_weights() like:
        >>> deepn4_model = DeepN4() # skip if not have_tf
        >>> deepn4_model.fetch_default_weights() # skip if not have_tf

        This model is designed to take as input file T1 signal and predict
        bias field. Effectively, this model is mimicking bias correction.

        Parameters
        ----------
        verbose : bool, optional
            Whether to show information about the processing.
        """
        if not have_tf:
            raise tf()

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)

        # Synb0 network load

        self.model = UNet3D(input_shape=(128, 128, 128, 1))

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting."""
        fetch_model_weights_path = get_fnames(name="deepn4_default_tf_weights")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (hdf5, saved by tensorflow)
        """
        try:
            self.model.load_weights(weights_path)
        except ValueError as e:
            raise ValueError(
                "Expected input for the provided model weights \
                             do not match the declared model"
            ) from e

    def __predict(self, x_test):
        """Internal prediction function
        Predict bias field from input T1 signal

        Parameters
        ----------
        x_test : np.ndarray (128, 128, 128, 1)
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (128, 128, 128)
            Predicted bias field
        """
        return self.model.predict(x_test)[..., 0]

    def pad(self, img, sz):
        tmp = np.zeros((sz, sz, sz))

        diff = int((sz - img.shape[0]) / 2)
        lx = max(diff, 0)
        lX = min(img.shape[0] + diff, sz)

        diff = (img.shape[0] - sz) / 2
        rx = max(int(np.floor(diff)), 0)
        rX = min(img.shape[0] - int(np.ceil(diff)), img.shape[0])

        diff = int((sz - img.shape[1]) / 2)
        ly = max(diff, 0)
        lY = min(img.shape[1] + diff, sz)

        diff = (img.shape[1] - sz) / 2
        ry = max(int(np.floor(diff)), 0)
        rY = min(img.shape[1] - int(np.ceil(diff)), img.shape[1])

        diff = int((sz - img.shape[2]) / 2)
        lz = max(diff, 0)
        lZ = min(img.shape[2] + diff, sz)

        diff = (img.shape[2] - sz) / 2
        rz = max(int(np.floor(diff)), 0)
        rZ = min(img.shape[2] - int(np.ceil(diff)), img.shape[2])

        tmp[lx:lX, ly:lY, lz:lZ] = img[rx:rX, ry:rY, rz:rZ]

        return tmp, [lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ]

    def load_resample(self, subj):
        input_data, [lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ] = self.pad(
            subj, 128
        )
        in_max = np.percentile(input_data[np.nonzero(input_data)], 99.99)
        input_data = normalize(input_data, min_v=0, max_v=in_max, new_min=0, new_max=1)
        input_data = np.squeeze(input_data)
        input_vols = np.zeros((1, 128, 128, 128, 1))
        input_vols[0, :, :, :, 0] = input_data

        return (
            tf.convert_to_tensor(input_vols, dtype=tf.float32),
            lx,
            lX,
            ly,
            lY,
            lz,
            lZ,
            rx,
            rX,
            ry,
            rY,
            rz,
            rZ,
            in_max,
        )

    def predict(self, img, img_affine, *, voxsize=(1, 1, 1), threshold=0.5):
        """Wrapper function to facilitate prediction of larger dataset.
        The function will mask, normalize, split, predict and 're-assemble'
        the data as a volume.

        Parameters
        ----------
        img : np.ndarray
            T1 image to predict and apply bias field
        img_affine : np.ndarray (4, 4)
            Affine matrix for the T1 image
        voxsize : np.ndarray or list or tuple (3,), optional
            voxel size of the T1 image.
        threshold : float, optional
            Threshold for cleaning the final correction field

        Returns
        -------
        final_corrected : np.ndarray (x, y, z)
            Predicted bias corrected image.
            The volume has matching shape to the input data
        """
        # Preprocess input data (resample, normalize, and pad)
        resampled_T1, inv_affine, mid_shape, offset_array, scale, crop_vs, pad_vs = (
            transform_img(img, img_affine, voxsize=voxsize)
        )
        (in_features, lx, lX, ly, lY, lz, lZ, rx, rX, ry, rY, rz, rZ, in_max) = (
            self.load_resample(resampled_T1)
        )

        # Run the model to get the bias field
        logfield = self.__predict(in_features)
        field = np.exp(logfield)
        field = field.squeeze()

        # Postprocess predicted field (reshape - unpad, smooth the field,
        # upsample)
        final_field = np.zeros(
            [resampled_T1.shape[0], resampled_T1.shape[1], resampled_T1.shape[2]]
        )
        final_field[rx:rX, ry:rY, rz:rZ] = field[lx:lX, ly:lY, lz:lZ]
        final_fields = gaussian_filter(final_field, sigma=3)
        upsample_final_field = recover_img(
            final_fields,
            inv_affine,
            mid_shape,
            img.shape,
            offset_array,
            voxsize,
            scale,
            crop_vs,
            pad_vs,
        )

        # Correct the image
        below_threshold_mask = np.abs(upsample_final_field) < threshold
        with np.errstate(divide="ignore", invalid="ignore"):
            final_corrected = np.where(
                below_threshold_mask, 0, img / upsample_final_field
            )

        return final_corrected



import numpy as np
import numpy.testing as npt

from dipy.direction.peaks import PeaksAndMetrics
from dipy.testing.decorators import set_random_number_generator
from dipy.viz.horizon.util import (
    check_img_dtype,
    check_img_shapes,
    check_peak_size,
    show_ellipsis,
    unpack_surface,
)


@set_random_number_generator()
def test_check_img_shapes(rng):
    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = 255 * rng.random((197, 233, 189))
    data1 = 255 * rng.random((197, 233, 189))
    images = [(data, affine), (data1, affine)]
    npt.assert_equal(check_img_shapes(images), (True, False))

    data1 = 255 * rng.random((200, 233, 189))
    images = [(data, affine), (data1, affine)]
    npt.assert_equal(check_img_shapes(images), (False, False))

    data = 255 * rng.random((197, 233, 189, 10))
    data1 = 255 * rng.random((197, 233, 189))
    images = [(data, affine), (data1, affine)]

    npt.assert_equal(check_img_shapes(images), (True, True))

    data = 255 * rng.random((197, 233, 189, 15))
    data1 = 255 * rng.random((197, 233, 189, 15))
    images = [(data, affine), (data1, affine)]

    npt.assert_equal(check_img_shapes(images), (True, True))

    data = 255 * rng.random((198, 233, 189, 14))
    data1 = 255 * rng.random((198, 233, 189, 15))
    images = [(data, affine), (data1, affine)]

    npt.assert_equal(check_img_shapes(images), (True, False))

    data = 255 * rng.random((197, 233, 189, 15))
    data1 = 255 * rng.random((198, 233, 189, 14))
    images = [(data, affine), (data1, affine)]

    npt.assert_equal(check_img_shapes(images), (False, False))

    data = 255 * rng.random((197, 233, 189, 15))
    data1 = 255 * rng.random((198, 233, 189, 15))
    images = [(data, affine), (data1, affine)]

    npt.assert_equal(check_img_shapes(images), (False, False))


@set_random_number_generator()
def test_check_img_dtype(rng):
    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = 255 * rng.random((197, 233, 189))
    images = [
        (data, affine),
    ]

    npt.assert_equal(check_img_dtype(images)[0], images[0])

    data = rng.random((5, 5, 5)).astype(np.int64)
    images = [
        (data, affine),
    ]
    npt.assert_equal(check_img_dtype(images)[0][0].dtype, np.int32)
    assert len(check_img_dtype(images)) == 1

    data = rng.random((5, 5, 5)).astype(np.float16)
    images = [
        (data, affine),
    ]
    npt.assert_equal(check_img_dtype(images)[0][0].dtype, np.float32)
    assert len(check_img_dtype(images)) == 1

    data = rng.random((5, 5, 5)).astype(np.bool_)
    images = [
        (data, affine),
    ]
    assert len(check_img_dtype(images)) == 0


def test_show_ellipsis():
    text = "IAmALongFileName"
    text_size = 10
    available_size = 5
    result_text = f"...{text[-5:]}"

    npt.assert_equal(show_ellipsis(text, text_size, available_size), result_text)

    available_size = 12
    npt.assert_equal(show_ellipsis(text, text_size, available_size), text)


@set_random_number_generator()
def test_unpack_surface(rng):
    vertices = rng.random((100, 4))
    faces = rng.integers(0, 100, size=(100, 3))

    with npt.assert_raises(ValueError):
        unpack_surface((vertices, faces))

    vertices = rng.random((100, 3))
    faces = rng.integers(0, 100, size=(100, 4))

    with npt.assert_raises(ValueError):
        unpack_surface((vertices, faces, "/test/filename.pial"))

    vertices = rng.random((100, 3))
    faces = rng.integers(0, 100, size=(100, 3))
    v, f, fname = unpack_surface((vertices, faces, "/test/filename.pial"))
    npt.assert_equal(vertices, v)
    npt.assert_equal(faces, f)
    npt.assert_equal("/test/filename.pial", fname)


@set_random_number_generator()
def test_check_peak_size(rng):
    peak_dirs = rng.random((100, 100, 100, 10, 6))

    pam = PeaksAndMetrics()
    pam.peak_dirs = peak_dirs

    npt.assert_equal(True, check_peak_size([pam]))
    npt.assert_equal(True, check_peak_size([pam, pam]))
    npt.assert_equal(
        False, check_peak_size([pam], ref_img_shape=(100, 100, 1), sync_imgs=True)
    )
    npt.assert_equal(
        False, check_peak_size([pam], ref_img_shape=(100, 100, 100), sync_imgs=False)
    )

    pam1 = PeaksAndMetrics()
    peak_dirs_1 = rng.random((100, 100, 50, 10, 6))
    pam1.peak_dirs = peak_dirs_1

    npt.assert_equal(False, check_peak_size([pam, pam1]))
    npt.assert_equal(
        False,
        check_peak_size([pam, pam1], ref_img_shape=(100, 100, 100), sync_imgs=True),
    )
    npt.assert_equal(
        False,
        check_peak_size([pam, pam1], ref_img_shape=(100, 100, 100), sync_imgs=False),
    )



from math import factorial
from warnings import warn

import numpy as np
from scipy.special import gamma, genlaguerre, hyp2f1

from dipy.core.geometry import cart2sphere
from dipy.reconst.cache import Cache
from dipy.reconst.multi_voxel import multi_voxel_fit
from dipy.reconst.shm import real_sh_descoteaux_from_index
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")


class ShoreModel(Cache):
    r"""Simple Harmonic Oscillator based Reconstruction and Estimation
    (SHORE) of the diffusion signal.

    The main idea of SHORE :footcite:p:`Ozarslan2008` is to model the diffusion
    signal as a linear combination of continuous functions $\phi_i$,

    .. math::

       S(\mathbf{q})= \sum_{i=0}^I  c_{i} \phi_{i}(\mathbf{q})

    where $\mathbf{q}$ is the wave vector which corresponds to different
    gradient directions. Numerous continuous functions $\phi_i$ can be used to
    model $S$. Some are presented in :footcite:p:`Merlet2013`,
    :footcite:p:`Rathi2011`, and :footcite:p:`Cheng2011`.

    From the $c_i$ coefficients, there exist analytical formulae to estimate
    the ODF, the return to the origin probability (RTOP), the mean square
    displacement (MSD), amongst others :footcite:p:`Ozarslan2013`.

    References
    ----------
    .. footbibliography::

    Notes
    -----
    The implementation of SHORE depends on CVXPY (https://www.cvxpy.org/).
    """

    @warning_for_keywords()
    def __init__(
        self,
        gtab,
        *,
        radial_order=6,
        zeta=700,
        lambdaN=1e-8,
        lambdaL=1e-8,
        tau=1.0 / (4 * np.pi**2),
        constrain_e0=False,
        positive_constraint=False,
        pos_grid=11,
        pos_radius=20e-03,
        cvxpy_solver=None,
    ):
        r"""Analytical and continuous modeling of the diffusion signal with
        respect to the SHORE basis.

        This implementation is a modification of SHORE presented in
        :footcite:p:`Merlet2013`. The modification was made to obtain the same
        ordering of the basis presented in :footcite:p:`Cheng2011`,
        :footcite:p:`Ozarslan2013`.

        The main idea is to model the diffusion signal as a linear
        combination of continuous functions $\phi_i$,

        .. math::

           S(\mathbf{q})= \sum_{i=0}^I  c_{i} \phi_{i}(\mathbf{q})

        where $\mathbf{q}$ is the wave vector which corresponds to different
        gradient directions.

        From the $c_i$ coefficients, there exists an analytical formula to
        estimate the ODF.

        Parameters
        ----------
        gtab : GradientTable,
            gradient directions and bvalues container class
        radial_order : unsigned int, optional
            an even integer that represent the order of the basis
        zeta : unsigned int, optional
            scale factor
        lambdaN : float, optional
            radial regularisation constant
        lambdaL : float, optional
            angular regularisation constant
        tau : float, optional
            diffusion time. By default the value that makes q equal to the
            square root of the b-value.
        constrain_e0 : bool, optional
            Constrain the optimization such that E(0) = 1.
        positive_constraint : bool, optional
            Constrain the propagator to be positive.
        pos_grid : int, optional
            Grid that define the points of the EAP in which we want to enforce
            positivity.
        pos_radius : float, optional
            Radius of the grid of the EAP in which enforce positivity in
            millimeters.
        cvxpy_solver : str, optional
            cvxpy solver name. Optionally optimize the positivity constraint
            with a particular cvxpy solver. See https://www.cvxpy.org/ for
            details.
            Default: None (cvxpy chooses its own solver)

        References
        ----------
        .. footbibliography::

        Examples
        --------
        In this example, where the data, gradient table and sphere tessellation
        used for reconstruction are provided, we model the diffusion signal
        with respect to the SHORE basis and compute the real and analytical
        ODF.

        >>> import warnings
        >>> from dipy.data import get_isbi2013_2shell_gtab, default_sphere
        >>> from dipy.sims.voxel import sticks_and_ball
        >>> from dipy.reconst.shm import descoteaux07_legacy_msg
        >>> from dipy.reconst.shore import ShoreModel
        >>> gtab = get_isbi2013_2shell_gtab()
        >>> data, golden_directions = sticks_and_ball(
        ...    gtab, d=0.0015, S0=1., angles=[(0, 0), (90, 0)],
        ...    fractions=[50, 50], snr=None)
        ...
        >>> radial_order = 4
        >>> zeta = 700
        >>> asm = ShoreModel(gtab, radial_order=radial_order, zeta=zeta,
        ...                  lambdaN=1e-8, lambdaL=1e-8)
        >>> with warnings.catch_warnings():
        ...     warnings.filterwarnings(
        ...         "ignore", message=descoteaux07_legacy_msg,
        ...         category=PendingDeprecationWarning)
        ...     asmfit = asm.fit(data)
        ...     odf = asmfit.odf(default_sphere)
        """

        self.bvals = gtab.bvals
        self.bvecs = gtab.bvecs
        self.gtab = gtab
        self.constrain_e0 = constrain_e0
        if radial_order > 0 and not (bool(radial_order % 2)):
            self.radial_order = radial_order
        else:
            msg = "radial_order must be a non-zero even positive number."
            raise ValueError(msg)
        self.zeta = zeta
        self.lambdaL = lambdaL
        self.lambdaN = lambdaN
        if (gtab.big_delta is None) or (gtab.small_delta is None):
            self.tau = tau
        else:
            self.tau = gtab.big_delta - gtab.small_delta / 3.0

        if positive_constraint and not constrain_e0:
            msg = "Constrain_e0 must be True to enforce positivity."
            raise ValueError(msg)

        if positive_constraint or constrain_e0:
            if not have_cvxpy:
                msg = "cvxpy must be installed for positive_constraint or "
                msg += "constraint_e0."
                raise ImportError(msg)
            if cvxpy_solver is not None:
                if cvxpy_solver not in cvxpy.installed_solvers():
                    msg = f"Input `cvxpy_solver` was set to {cvxpy_solver}."
                    msg += f" One of {', '.join(cvxpy.installed_solvers())}"
                    msg += " was expected."
                    raise ValueError(msg)

        self.cvxpy_solver = cvxpy_solver
        self.positive_constraint = positive_constraint
        self.pos_grid = pos_grid
        self.pos_radius = pos_radius

    @multi_voxel_fit
    def fit(self, data, **kwargs):
        Lshore = l_shore(self.radial_order)
        Nshore = n_shore(self.radial_order)
        # Generate the SHORE basis
        M = self.cache_get("shore_matrix", key=self.gtab)
        if M is None:
            M = shore_matrix(self.radial_order, self.zeta, self.gtab, tau=self.tau)
            self.cache_set("shore_matrix", self.gtab, M)

        MpseudoInv = self.cache_get("shore_matrix_reg_pinv", key=self.gtab)
        if MpseudoInv is None:
            MpseudoInv = np.dot(
                np.linalg.inv(
                    np.dot(M.T, M) + self.lambdaN * Nshore + self.lambdaL * Lshore
                ),
                M.T,
            )
            self.cache_set("shore_matrix_reg_pinv", self.gtab, MpseudoInv)

        # Compute the signal coefficients in SHORE basis
        if not self.constrain_e0:
            coef = np.dot(MpseudoInv, data)

            signal_0 = 0

            for n in range(int(self.radial_order / 2) + 1):
                signal_0 += coef[n] * (
                    genlaguerre(n, 0.5)(0)
                    * ((factorial(n)) / (2 * np.pi * (self.zeta**1.5) * gamma(n + 1.5)))
                    ** 0.5
                )

            coef = coef / signal_0
        else:
            data_norm = data / data[self.gtab.b0s_mask].mean()
            M0 = M[self.gtab.b0s_mask, :]

            c = cvxpy.Variable(M.shape[1])
            design_matrix = cvxpy.Constant(M) @ c
            objective = cvxpy.Minimize(
                cvxpy.sum_squares(design_matrix - data_norm)
                + self.lambdaN * cvxpy.quad_form(c, Nshore)
                + self.lambdaL * cvxpy.quad_form(c, Lshore)
            )

            if not self.positive_constraint:
                constraints = [M0[0] @ c == 1]
            else:
                lg = int(np.floor(self.pos_grid**3 / 2))
                v, t = create_rspace(self.pos_grid, self.pos_radius)
                psi = self.cache_get(
                    "shore_matrix_positive_constraint",
                    key=(self.pos_grid, self.pos_radius),
                )
                if psi is None:
                    psi = shore_matrix_pdf(self.radial_order, self.zeta, t[:lg])
                    self.cache_set(
                        "shore_matrix_positive_constraint",
                        (self.pos_grid, self.pos_radius),
                        psi,
                    )
                constraints = [(M0[0] @ c) == 1.0, (psi @ c) >= 1e-3]
            prob = cvxpy.Problem(objective, constraints)
            try:
                prob.solve(solver=self.cvxpy_solver)
                coef = np.asarray(c.value).squeeze()
            except Exception:
                warn("Optimization did not find a solution", stacklevel=2)
                coef = np.zeros(M.shape[1])
        return ShoreFit(self, coef)


class ShoreFit:
    def __init__(self, model, shore_coef):
        """Calculates diffusion properties for a single voxel

        Parameters
        ----------
        model : object,
            AnalyticalModel
        shore_coef : 1d ndarray,
            shore coefficients
        """

        self.model = model
        self._shore_coef = shore_coef
        self.gtab = model.gtab
        self.radial_order = model.radial_order
        self.zeta = model.zeta

    def pdf_grid(self, gridsize, radius_max):
        r"""Applies the analytical FFT on $S$ to generate the diffusion
        propagator. This is calculated on a discrete 3D grid in order to
        obtain an EAP similar to that which is obtained with DSI.

        Parameters
        ----------
        gridsize : unsigned int
            dimension of the propagator grid
        radius_max : float
            maximal radius in which to compute the propagator

        Returns
        -------
        eap : ndarray
            the ensemble average propagator in the 3D grid

        """
        # Create the grid in which to compute the pdf
        rgrid_rtab = self.model.cache_get("pdf_grid", key=(gridsize, radius_max))
        if rgrid_rtab is None:
            rgrid_rtab = create_rspace(gridsize, radius_max)
            self.model.cache_set("pdf_grid", (gridsize, radius_max), rgrid_rtab)
        rgrid, rtab = rgrid_rtab

        psi = self.model.cache_get("shore_matrix_pdf", key=(gridsize, radius_max))
        if psi is None:
            psi = shore_matrix_pdf(self.radial_order, self.zeta, rtab)
            self.model.cache_set("shore_matrix_pdf", (gridsize, radius_max), psi)

        propagator = np.dot(psi, self._shore_coef)
        eap = np.empty((gridsize, gridsize, gridsize), dtype=float)
        eap[tuple(rgrid.astype(int).T)] = propagator
        eap *= (2 * radius_max / (gridsize - 1)) ** 3

        return eap

    def pdf(self, r_points):
        """Diffusion propagator on a given set of real points.
        if the array r_points is non writeable, then intermediate
        results are cached for faster recalculation
        """
        if not r_points.flags.writeable:
            psi = self.model.cache_get("shore_matrix_pdf", key=hash(r_points.data))
        else:
            psi = None
        if psi is None:
            psi = shore_matrix_pdf(self.radial_order, self.zeta, r_points)
            if not r_points.flags.writeable:
                self.model.cache_set("shore_matrix_pdf", hash(r_points.data), psi)

        eap = np.dot(psi, self._shore_coef)

        return np.clip(eap, 0, eap.max())

    def odf_sh(self):
        r"""Calculates the real analytical ODF in terms of Spherical
        Harmonics.
        """
        # Number of Spherical Harmonics involved in the estimation
        J = (self.radial_order + 1) * (self.radial_order + 2) // 2

        # Compute the Spherical Harmonics Coefficients
        c_sh = np.zeros(J)
        counter = 0

        for ell in range(0, self.radial_order + 1, 2):
            for n in range(ell, int((self.radial_order + ell) / 2) + 1):
                for m in range(-ell, ell + 1):
                    j = int(ell + m + (2 * np.array(range(0, ell, 2)) + 1).sum())

                    Cnl = (
                        ((-1) ** (n - ell / 2))
                        / (2.0 * (4.0 * np.pi**2 * self.zeta) ** (3.0 / 2.0))
                        * (
                            (
                                2.0
                                * (4.0 * np.pi**2 * self.zeta) ** (3.0 / 2.0)
                                * factorial(n - ell)
                            )
                            / (gamma(n + 3.0 / 2.0))
                        )
                        ** (1.0 / 2.0)
                    )
                    Gnl = (
                        (gamma(ell / 2 + 3.0 / 2.0) * gamma(3.0 / 2.0 + n))
                        / (gamma(ell + 3.0 / 2.0) * factorial(n - ell))
                        * (1.0 / 2.0) ** (-ell / 2 - 3.0 / 2.0)
                    )
                    Fnl = hyp2f1(-n + ell, ell / 2 + 3.0 / 2.0, ell + 3.0 / 2.0, 2.0)

                    c_sh[j] += self._shore_coef[counter] * Cnl * Gnl * Fnl
                    counter += 1

        return c_sh

    def odf(self, sphere):
        r"""Calculates the ODF for a given discrete sphere."""
        upsilon = self.model.cache_get("shore_matrix_odf", key=sphere)
        if upsilon is None:
            upsilon = shore_matrix_odf(self.radial_order, self.zeta, sphere.vertices)
            self.model.cache_set("shore_matrix_odf", sphere, upsilon)

        odf = np.dot(upsilon, self._shore_coef)
        return odf

    def rtop_signal(self):
        r"""Calculates the analytical return to origin probability (RTOP)
        from the signal.

        See :footcite:p:`Ozarslan2013` for further details about the method.

        References
        ----------
        .. footbibliography::
        """
        rtop = 0
        c = self._shore_coef

        for n in range(int(self.radial_order / 2) + 1):
            rtop += (
                c[n]
                * (-1) ** n
                * ((16 * np.pi * self.zeta**1.5 * gamma(n + 1.5)) / (factorial(n)))
                ** 0.5
            )

        return np.clip(rtop, 0, rtop.max())

    def rtop_pdf(self):
        r"""Calculates the analytical return to origin probability (RTOP)
        from the pdf.

        See :footcite:p:`Ozarslan2013` for further details about the method.

        References
        ----------
        .. footbibliography::
        """
        rtop = 0
        c = self._shore_coef
        for n in range(int(self.radial_order / 2) + 1):
            rtop += (
                c[n]
                * (-1) ** n
                * ((4 * np.pi**2 * self.zeta**1.5 * factorial(n)) / (gamma(n + 1.5)))
                ** 0.5
                * genlaguerre(n, 0.5)(0)
            )

        return np.clip(rtop, 0, rtop.max())

    def msd(self):
        r"""Calculates the analytical mean squared displacement (MSD).

        See :footcite:p:`Wu2007` for a definition of the method.

        .. math::
           :nowrap:

            MSD:{DSI}=\int_{-\infty}^{\infty}\int_{-\infty}^{\infty}
            \int_{-\infty}^{\infty} P(\hat{\mathbf{r}}) \cdot
            \hat{\mathbf{r}}^{2} \ dr_x \ dr_y \ dr_z

        where $\hat{\mathbf{r}}$ is a point in the 3D propagator space (see
        :footcite:t:`Wu2007`).

        References
        ----------
        .. footbibliography::
        """
        msd = 0
        c = self._shore_coef

        for n in range(int(self.radial_order / 2) + 1):
            msd += (
                c[n]
                * (-1) ** n
                * (
                    9
                    * (gamma(n + 1.5))
                    / (8 * np.pi**6 * self.zeta**3.5 * factorial(n))
                )
                ** 0.5
                * hyp2f1(-n, 2.5, 1.5, 2)
            )

        return np.clip(msd, 0, msd.max())

    def fitted_signal(self):
        """The fitted signal."""
        phi = self.model.cache_get("shore_matrix", key=self.model.gtab)
        return np.dot(phi, self._shore_coef)

    @property
    def shore_coeff(self):
        """The SHORE coefficients"""
        return self._shore_coef


@warning_for_keywords()
def shore_matrix(radial_order, zeta, gtab, *, tau=1 / (4 * np.pi**2)):
    r"""Compute the SHORE matrix for modified Merlet's 3D-SHORE.

    See :footcite:p:`Merlet2013` for the definition.

    .. math::
        :nowrap:

        \textbf{E}(q\textbf{u})=\sum_{l=0, even}^{N_{max}}
                               \sum_{n=l}^{(N_{max}+l)/2}
                               \sum_{m=-l}^l c_{nlm}
                               \phi_{nlm}(q\textbf{u})

    where $\phi_{nlm}$ is

    .. math::
        :nowrap:

        \phi_{nlm}^{SHORE}(q\textbf{u})=\Biggl[\dfrac{2(n-l)!}
            {\zeta^{3/2} \Gamma(n+3/2)} \Biggr]^{1/2}
            \Biggl(\dfrac{q^2}{\zeta}\Biggr)^{l/2}
            exp\Biggl(\dfrac{-q^2}{2\zeta}\Biggr)
            L^{l+1/2}_{n-l} \Biggl(\dfrac{q^2}{\zeta}\Biggr)
            Y_l^m(\textbf{u})

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer that represent the order of the basis
    zeta : unsigned int,
        scale factor
    gtab : GradientTable,
        gradient directions and bvalues container class
    tau : float, optional
        diffusion time. By default the value that makes q=sqrt(b).

    References
    ----------
    .. footbibliography::

    """

    qvals = np.sqrt(gtab.bvals / (4 * np.pi**2 * tau))
    qvals[gtab.b0s_mask] = 0
    bvecs = gtab.bvecs

    qgradients = qvals[:, None] * bvecs

    r, theta, phi = cart2sphere(qgradients[:, 0], qgradients[:, 1], qgradients[:, 2])
    theta[np.isnan(theta)] = 0
    F = radial_order / 2
    n_c = int(np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3)))
    M = np.zeros((r.shape[0], n_c))

    counter = 0
    for ell in range(0, radial_order + 1, 2):
        for n in range(ell, int((radial_order + ell) / 2) + 1):
            for m in range(-ell, ell + 1):
                M[:, counter] = (
                    real_sh_descoteaux_from_index(m, ell, theta, phi)
                    * genlaguerre(n - ell, ell + 0.5)(r**2 / zeta)
                    * np.exp(-(r**2) / (2.0 * zeta))
                    * _kappa(zeta, n, ell)
                    * (r**2 / zeta) ** (ell / 2)
                )
                counter += 1
    return M


def _kappa(zeta, n, ell):
    return np.sqrt((2 * factorial(n - ell)) / (zeta**1.5 * gamma(n + 1.5)))


def shore_matrix_pdf(radial_order, zeta, rtab):
    r"""Compute the SHORE propagator matrix.

    See :footcite:p:`Merlet2013` for the definition.

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer that represent the order of the basis
    zeta : unsigned int,
        scale factor
    rtab : array, shape (N,3)
        real space points in which calculates the pdf

    References
    ----------
    .. footbibliography::
    """

    r, theta, phi = cart2sphere(rtab[:, 0], rtab[:, 1], rtab[:, 2])
    theta[np.isnan(theta)] = 0
    F = radial_order / 2
    n_c = int(np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3)))
    psi = np.zeros((r.shape[0], n_c))
    counter = 0
    for ell in range(0, radial_order + 1, 2):
        for n in range(ell, int((radial_order + ell) / 2) + 1):
            for m in range(-ell, ell + 1):
                psi[:, counter] = (
                    real_sh_descoteaux_from_index(m, ell, theta, phi)
                    * genlaguerre(n - ell, ell + 0.5)(4 * np.pi**2 * zeta * r**2)
                    * np.exp(-2 * np.pi**2 * zeta * r**2)
                    * _kappa_pdf(zeta, n, ell)
                    * (4 * np.pi**2 * zeta * r**2) ** (ell / 2)
                    * (-1) ** (n - ell / 2)
                )
                counter += 1
    return psi


def _kappa_pdf(zeta, n, ell):
    return np.sqrt((16 * np.pi**3 * zeta**1.5 * factorial(n - ell)) / gamma(n + 1.5))


def shore_matrix_odf(radial_order, zeta, sphere_vertices):
    r"""Compute the SHORE ODF matrix.

    See :footcite:p:`Merlet2013` for the definition.

    Parameters
    ----------
    radial_order : unsigned int,
        an even integer that represent the order of the basis
    zeta : unsigned int,
        scale factor
    sphere_vertices : array, shape (N,3)
        vertices of the odf sphere

    References
    ----------
    .. footbibliography::
    """

    r, theta, phi = cart2sphere(
        sphere_vertices[:, 0], sphere_vertices[:, 1], sphere_vertices[:, 2]
    )
    theta[np.isnan(theta)] = 0
    F = radial_order / 2
    n_c = int(np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3)))
    upsilon = np.zeros((len(sphere_vertices), n_c))
    counter = 0
    for ell in range(0, radial_order + 1, 2):
        for n in range(ell, int((radial_order + ell) / 2) + 1):
            for m in range(-ell, ell + 1):
                upsilon[:, counter] = (
                    (-1) ** (n - ell / 2.0)
                    * _kappa_odf(zeta, n, ell)
                    * hyp2f1(ell - n, ell / 2.0 + 1.5, ell + 1.5, 2.0)
                    * real_sh_descoteaux_from_index(m, ell, theta, phi)
                )
                counter += 1

    return upsilon


def _kappa_odf(zeta, n, ell):
    return np.sqrt(
        (gamma(ell / 2.0 + 1.5) ** 2 * gamma(n + 1.5) * 2 ** (ell + 3))
        / (16 * np.pi**3 * zeta**1.5 * factorial(n - ell) * gamma(ell + 1.5) ** 2)
    )


def l_shore(radial_order):
    """Returns the angular regularisation matrix for SHORE basis"""
    F = radial_order / 2
    n_c = int(np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3)))
    diagL = np.zeros(n_c)
    counter = 0
    for ell in range(0, radial_order + 1, 2):
        for _ in range(ell, int((radial_order + ell) / 2) + 1):
            for _ in range(-ell, ell + 1):
                diagL[counter] = (ell * (ell + 1)) ** 2
                counter += 1

    return np.diag(diagL)


def n_shore(radial_order):
    """Returns the angular regularisation matrix for SHORE basis"""
    F = radial_order / 2
    n_c = int(np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3)))
    diagN = np.zeros(n_c)
    counter = 0
    for ell in range(0, radial_order + 1, 2):
        for n in range(ell, int((radial_order + ell) / 2) + 1):
            for _ in range(-ell, ell + 1):
                diagN[counter] = (n * (n + 1)) ** 2
                counter += 1

    return np.diag(diagN)


def create_rspace(gridsize, radius_max):
    """Create the real space table, that contains the points in which
        to compute the pdf.

    Parameters
    ----------
    gridsize : unsigned int
        dimension of the propagator grid
    radius_max : float
        maximal radius in which compute the propagator

    Returns
    -------
    vecs : array, shape (N,3)
        positions of the pdf points in a 3D matrix

    tab : array, shape (N,3)
        real space points in which calculates the pdf
    """

    radius = gridsize // 2
    vecs = []
    for i in range(-radius, radius + 1):
        for j in range(-radius, radius + 1):
            for k in range(-radius, radius + 1):
                vecs.append([i, j, k])

    vecs = np.array(vecs, dtype=np.float32)
    tab = vecs / radius
    tab = tab * radius_max
    vecs = vecs + radius

    return vecs, tab


def shore_indices(radial_order, index):
    r"""Given the basis order and the index, return the shore indices n, l, m
    for modified Merlet's 3D-SHORE

    .. math::
        :nowrap:

        \begin{equation}
            \textbf{E}(q\textbf{u})=\sum_{l=0, even}^{N_{max}}
                                    \sum_{n=l}^{(N_{max}+l)/2}
                                    \sum_{m=-l}^l c_{nlm}
                                    \phi_{nlm}(q\textbf{u})
        \end{equation}

    where $\phi_{nlm}$ is

    .. math::
        :nowrap:

        \begin{equation}
            \phi_{nlm}^{SHORE}(q\textbf{u})=\Biggl[\dfrac{2(n-l)!}
                {\zeta^{3/2} \Gamma(n+3/2)} \Biggr]^{1/2}
                \Biggl(\dfrac{q^2}{\zeta}\Biggr)^{l/2}
                exp\Biggl(\dfrac{-q^2}{2\zeta}\Biggr)
                L^{l+1/2}_{n-l} \Biggl(\dfrac{q^2}{\zeta}\Biggr)
                Y_l^m(\textbf{u}).
        \end{equation}

    Parameters
    ----------
    radial_order : unsigned int
        an even integer that represent the maximal order of the basis
    index : unsigned int
        index of the coefficients, start from 0

    Returns
    -------
    n :  unsigned int
        the index n of the modified shore basis
    l :  unsigned int
        the index l of the modified shore basis
    m :  unsigned int
        the index m of the modified shore basis
    """

    F = radial_order / 2
    n_c = np.round(1 / 6.0 * (F + 1) * (F + 2) * (4 * F + 3))
    n_i = 0
    l_i = 0
    m_i = 0

    if n_c < (index + 1):
        msg = f"The index {index} is higher than the number of"
        msg += " coefficients of the truncated basis,"
        msg += f" which is {int(n_c - 1)} starting from 0."
        msg += " Select a lower index."
        raise ValueError(msg)
    else:
        counter = 0
        for ell in range(0, radial_order + 1, 2):
            for n in range(ell, int((radial_order + ell) / 2) + 1):
                for m in range(-ell, ell + 1):
                    if counter == index:
                        n_i = n
                        l_i = ell
                        m_i = m
                    counter += 1
    return n_i, l_i, m_i


def shore_order(n, ell, m):
    r"""Given the indices (n,l,m) of the basis, return the minimum order
    for those indices and their index for modified Merlet's 3D-SHORE.

    Parameters
    ----------
    n :  unsigned int
        the index n of the modified shore basis
    ell :  unsigned int
        the index l of the modified shore basis
    m :  unsigned int
        the index m of the modified shore basis

    Returns
    -------
    radial_order : unsigned int
        an even integer that represent the maximal order of the basis
    index : unsigned int
        index of the coefficient corresponding to (n,l,m), start from 0

    """
    if ell % 2 == 1 or ell > n or ell < 0 or n < 0 or np.abs(m) > ell:
        msg = "The index l must be even and 0 <= l <= n, the index m must be "
        msg += "-l <= m <= ell. Given values were"
        msg += f" [n,l,m]=[{','.join([str(n), str(ell), str(m)])}]."
        raise ValueError(msg)
    else:
        if n % 2 == 1:
            radial_order = n + 1
        else:
            radial_order = n

        counter_i = 0

        counter = 0
        for l_i in range(0, radial_order + 1, 2):
            for n_i in range(l_i, int((radial_order + l_i) / 2) + 1):
                for m_i in range(-l_i, l_i + 1):
                    if n == n_i and ell == l_i and m == m_i:
                        counter_i = counter
                    counter += 1

    return radial_order, counter_i



from nibabel.affines import voxel_sizes
import numpy as np

from dipy.data import default_sphere
from dipy.direction import (
    BootDirectionGetter,
    ClosestPeakDirectionGetter,
    ProbabilisticDirectionGetter,
)
from dipy.direction.peaks import peaks_from_positions
from dipy.direction.pmf import SHCoeffPmfGen, SimplePmfGen
from dipy.tracking.local_tracking import LocalTracking, ParticleFilteringTracking
from dipy.tracking.tracker_parameters import generate_tracking_parameters
from dipy.tracking.tractogen import generate_tractogram
from dipy.tracking.utils import seeds_directions_pairs


def generic_tracking(
    seed_positions,
    seed_directions,
    sc,
    params,
    *,
    affine=None,
    sh=None,
    peaks=None,
    sf=None,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    seed_buffer_fraction=1.0,
    save_seeds=False,
):
    affine = affine if affine is not None else np.eye(4)

    pmf_type = [
        {"name": "sh", "value": sh, "cls": SHCoeffPmfGen},
        {"name": "peaks", "value": peaks, "cls": SimplePmfGen},
        {"name": "sf", "value": sf, "cls": SimplePmfGen},
    ]

    initialized_pmf = [
        d_selected for d_selected in pmf_type if d_selected["value"] is not None
    ]
    if len(initialized_pmf) > 1:
        selected_pmf = ", ".join([p["name"] for p in initialized_pmf])
        raise ValueError(
            "Only one pmf type should be initialized. "
            f"Variables initialized: {', '.join(selected_pmf)}"
        )
    if len(initialized_pmf) == 0:
        available_pmf = ", ".join([d["name"] for d in pmf_type])
        raise ValueError(
            f"No PMF found. One of this variable ({available_pmf}) should be"
            " initialized."
        )

    selected_pmf = initialized_pmf[0]

    if selected_pmf["name"] == "sf" and sphere is None:
        raise ValueError("A sphere should be defined when using SF (an ODF).")

    if selected_pmf["name"] == "peaks":
        raise NotImplementedError("Peaks are not yet implemented.")

    sphere = sphere or default_sphere

    kwargs = {}
    if selected_pmf["name"] == "sh":
        kwargs = {"basis_type": basis_type, "legacy": legacy}

    pmf_gen = selected_pmf["cls"](
        np.asarray(selected_pmf["value"], dtype=float), sphere, **kwargs
    )

    if seed_directions is not None:
        if not isinstance(seed_directions, (np.ndarray, list)):
            raise ValueError("seed_directions should be a numpy array or a list.")
        elif isinstance(seed_directions, list):
            seed_directions = np.array(seed_directions)

        if not np.array_equal(seed_directions.shape, seed_positions.shape):
            raise ValueError(
                "seed_directions and seed_positions should have the same shape."
            )
    else:
        peaks = peaks_from_positions(
            seed_positions, None, None, npeaks=1, affine=affine, pmf_gen=pmf_gen
        )
        seed_positions, seed_directions = seeds_directions_pairs(
            seed_positions, peaks, max_cross=1
        )

    return generate_tractogram(
        seed_positions,
        seed_directions,
        sc,
        params,
        pmf_gen,
        affine=affine,
        nbr_threads=nbr_threads,
        buffer_frac=seed_buffer_fraction,
        save_seeds=save_seeds,
    )


def probabilistic_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    min_len=2,
    max_len=500,
    step_size=0.2,
    voxel_size=None,
    max_angle=20,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """Probabilistic tracking algorithm.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
       Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    min_len : int, optional
        Minimum length (Nb points) of the streamlines.
    max_len : int, optional
        Maximum length (Nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.

    Returns
    -------
    Tractogram

    """
    voxel_size = voxel_size if voxel_size is not None else voxel_sizes(affine)

    params = generate_tracking_parameters(
        "prob",
        min_len=min_len,
        max_len=max_len,
        step_size=step_size,
        voxel_size=voxel_size,
        max_angle=max_angle,
        pmf_threshold=pmf_threshold,
        random_seed=random_seed,
        return_all=return_all,
    )

    return generic_tracking(
        seed_positions,
        seed_directions,
        sc,
        params,
        affine=affine,
        sh=sh,
        peaks=peaks,
        sf=sf,
        sphere=sphere,
        basis_type=basis_type,
        legacy=legacy,
        nbr_threads=nbr_threads,
        seed_buffer_fraction=seed_buffer_fraction,
        save_seeds=save_seeds,
    )


def deterministic_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    min_len=2,
    max_len=500,
    step_size=0.2,
    voxel_size=None,
    max_angle=20,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """Deterministic tracking algorithm.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
        Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    min_len : int, optional
        Minimum length (nb points) of the streamlines.
    max_len : int, optional
        Maximum length (nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.

    Returns
    -------
    Tractogram

    """
    voxel_size = voxel_size if voxel_size is not None else voxel_sizes(affine)

    params = generate_tracking_parameters(
        "det",
        min_len=min_len,
        max_len=max_len,
        step_size=step_size,
        voxel_size=voxel_size,
        max_angle=max_angle,
        pmf_threshold=pmf_threshold,
        random_seed=random_seed,
        return_all=return_all,
    )
    return generic_tracking(
        seed_positions,
        seed_directions,
        sc,
        params,
        affine=affine,
        sh=sh,
        peaks=peaks,
        sf=sf,
        sphere=sphere,
        basis_type=basis_type,
        legacy=legacy,
        nbr_threads=nbr_threads,
        seed_buffer_fraction=seed_buffer_fraction,
        save_seeds=save_seeds,
    )


def ptt_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    min_len=2,
    max_len=500,
    step_size=0.5,
    voxel_size=None,
    max_angle=10,
    pmf_threshold=0.1,
    probe_length=1.5,
    probe_radius=0,
    probe_quality=7,
    probe_count=1,
    data_support_exponent=1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """Parallel Transport Tractography (PTT) tracking algorithm.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
        Spherical Harmonics (SH) data.
    peaks : ndarray, optional
        Peaks array
    sf : ndarray, optional
        Spherical Function (SF).
    min_len : int, optional
        Minimum length of the streamlines.
    max_len : int, optional
        Maximum length of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    probe_length : float, optional
        Probe length.
    probe_radius : float, optional
        Probe radius.
    probe_quality : int, optional
        Probe quality.
    probe_count : int, optional
        Probe count.
    data_support_exponent : int, optional
        Data support exponent.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.
    Returns
    -------
    Tractogram

    """
    voxel_size = voxel_size if voxel_size is not None else voxel_sizes(affine)

    params = generate_tracking_parameters(
        "ptt",
        min_len=min_len,
        max_len=max_len,
        step_size=step_size,
        voxel_size=voxel_size,
        max_angle=max_angle,
        pmf_threshold=pmf_threshold,
        random_seed=random_seed,
        probe_length=probe_length,
        probe_radius=probe_radius,
        probe_quality=probe_quality,
        probe_count=probe_count,
        data_support_exponent=data_support_exponent,
        return_all=return_all,
    )
    return generic_tracking(
        seed_positions,
        seed_directions,
        sc,
        params,
        affine=affine,
        sh=sh,
        peaks=peaks,
        sf=sf,
        sphere=sphere,
        basis_type=basis_type,
        legacy=legacy,
        nbr_threads=nbr_threads,
        seed_buffer_fraction=seed_buffer_fraction,
        save_seeds=save_seeds,
    )


def closestpeak_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    min_len=2,
    max_len=500,
    step_size=0.5,
    voxel_size=None,
    max_angle=60,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """Closest peak tracking algorithm.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
        Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    min_len : int, optional
        Minimum length (nb points) of the streamlines.
    max_len : int, optional
        Maximum length (nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.

    Returns
    -------
    Tractogram

    """
    dg = None
    sphere = sphere if sphere is not None else default_sphere
    if sh is not None:
        dg = ClosestPeakDirectionGetter.from_shcoeff(
            sh,
            sphere=sphere,
            max_angle=max_angle,
            pmf_threshold=pmf_threshold,
            basis_type=basis_type,
            legacy=legacy,
        )
    elif sf is not None:
        dg = ClosestPeakDirectionGetter.from_pmf(
            sf, sphere=sphere, max_angle=max_angle, pmf_threshold=pmf_threshold
        )
    else:
        raise ValueError("SH or SF should be defined. Not implemented yet for peaks.")

    return LocalTracking(
        dg,
        sc,
        seed_positions,
        affine,
        step_size=step_size,
        minlen=min_len,
        maxlen=max_len,
        random_seed=random_seed,
        return_all=return_all,
        initial_directions=seed_directions,
        save_seeds=save_seeds,
    )


def bootstrap_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    data=None,
    model=None,
    sh=None,
    peaks=None,
    sf=None,
    min_len=2,
    max_len=500,
    step_size=0.5,
    voxel_size=None,
    max_angle=60,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """Bootstrap tracking algorithm.

    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    data : ndarray, optional
        Diffusion data.
    model : Model, optional
        Reconstruction model.
    sh : ndarray, optional
        Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    min_len : int, optional
        Minimum length (nb points) of the streamlines.
    max_len : int, optional
        Maximum length (nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.

    Returns
    -------
    Tractogram

    """
    sphere = sphere if sphere is not None else default_sphere
    if data is None or model is None:
        raise ValueError("Data and model should be defined.")

    dg = BootDirectionGetter.from_data(
        data,
        model,
        max_angle=max_angle,
    )
    return LocalTracking(
        dg,
        sc,
        seed_positions,
        affine,
        step_size=step_size,
        minlen=min_len,
        maxlen=max_len,
        random_seed=random_seed,
        return_all=return_all,
        initial_directions=seed_directions,
        save_seeds=save_seeds,
    )


def eudx_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    pam=None,
    min_len=2,
    max_len=500,
    step_size=0.5,
    voxel_size=None,
    max_angle=60,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    save_seeds=False,
):
    """EuDX tracking algorithm.

    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
        Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    pam : PeakAndMetrics, optional
        Peaks and Metrics object
    min_len : int, optional
        Minimum length (nb points) of the streamlines.
    max_len : int, optional
        Maximum length (nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.

    Returns
    -------
    Tractogram

    """
    sphere = sphere if sphere is not None else default_sphere
    if pam is None:
        raise ValueError("PAM should be defined.")

    return LocalTracking(
        pam,
        sc,
        seed_positions,
        affine,
        step_size=step_size,
        minlen=min_len,
        maxlen=max_len,
        random_seed=random_seed,
        return_all=return_all,
        initial_directions=seed_directions,
        save_seeds=save_seeds,
    )


def pft_tracking(
    seed_positions,
    sc,
    affine,
    *,
    seed_directions=None,
    sh=None,
    peaks=None,
    sf=None,
    pam=None,
    max_cross=None,
    min_len=2,
    max_len=500,
    step_size=0.2,
    voxel_size=None,
    max_angle=20,
    pmf_threshold=0.1,
    sphere=None,
    basis_type=None,
    legacy=True,
    nbr_threads=0,
    random_seed=0,
    seed_buffer_fraction=1.0,
    return_all=True,
    pft_back_tracking_dist=2,
    pft_front_tracking_dist=1,
    pft_max_trial=20,
    particle_count=15,
    save_seeds=False,
    min_wm_pve_before_stopping=0,
    unidirectional=False,
    randomize_forward_direction=False,
):
    """Particle Filtering Tracking (PFT) tracking algorithm.

    seed_positions : ndarray
        Seed positions in world space.
    sc : StoppingCriterion
        Stopping criterion.
    affine : ndarray
        Affine matrix.
    seed_directions : ndarray, optional
        Seed directions.
    sh : ndarray, optional
        Spherical Harmonics (SH).
    peaks : ndarray, optional
        Peaks array.
    sf : ndarray, optional
        Spherical Function (SF).
    pam : PeakAndMetrics, optional
        Peaks and Metrics object.
    max_cross : int, optional
        Maximum number of crossing fibers.
    min_len : int, optional
        Minimum length (nb points) of the streamlines.
    max_len : int, optional
        Maximum length (nb points) of the streamlines.
    step_size : float, optional
        Step size of the tracking.
    voxel_size : ndarray, optional
        Voxel size.
    max_angle : float, optional
        Maximum angle.
    pmf_threshold : float, optional
        PMF threshold.
    sphere : Sphere, optional
        Sphere.
    basis_type : name of basis
        The basis that ``shcoeff`` are associated with.
        ``dipy.reconst.shm.real_sh_descoteaux`` is used by default.
    legacy: bool, optional
        True to use a legacy basis definition for backward compatibility
        with previous ``tournier07`` and ``descoteaux07`` implementations.
    nbr_threads: int, optional
        Number of threads to use for the processing. By default, all available threads
        will be used.
    random_seed: int, optional
        Seed for the random number generator, must be >= 0. A value of greater than 0
        will all produce the same streamline trajectory for a given seed coordinate.
        A value of 0 may produces various streamline tracjectories for a given seed
        coordinate.
    seed_buffer_fraction: float, optional
        Fraction of the seed buffer to use. A value of 1.0 will use the entire seed
        buffer. A value of 0.5 will use half of the seed buffer then the other half.
        a way to reduce memory usage.
    return_all: bool, optional
        True to return all the streamlines, False to return only the streamlines that
        reached the stopping criterion.
    pft_back_tracking_dist : float, optional
        Back tracking distance.
    pft_front_tracking_dist : float, optional
        Front tracking distance.
    pft_max_trial : int, optional
        Maximum number of trials.
    particle_count : int, optional
        Number of particles.
    save_seeds: bool, optional
        True to return the seeds with the associated streamline.
    min_wm_pve_before_stopping : float, optional
        Minimum white matter partial volume estimation before stopping.
    unidirectional : bool, optional
        True to use unidirectional tracking.
    randomize_forward_direction : bool, optional
        True to randomize forward direction

    Returns
    -------
    Tractogram

    """
    sphere = sphere if sphere is not None else default_sphere

    dg = None
    if sh is not None:
        dg = ProbabilisticDirectionGetter.from_shcoeff(
            sh,
            max_angle=max_angle,
            sphere=sphere,
            sh_to_pmf=True,
            pmf_threshold=pmf_threshold,
            basis_type=basis_type,
            legacy=legacy,
        )
    elif sf is not None:
        dg = ProbabilisticDirectionGetter.from_pmf(
            sf, max_angle=max_angle, sphere=sphere, pmf_threshold=pmf_threshold
        )
    elif pam is not None and sh is None:
        sh = pam.shm_coeff
    else:
        msg = "SH, SF or PAM should be defined. Not implemented yet for peaks."
        raise ValueError(msg)

    return ParticleFilteringTracking(
        dg,
        sc,
        seed_positions,
        affine,
        max_cross=max_cross,
        step_size=step_size,
        minlen=min_len,
        maxlen=max_len,
        pft_back_tracking_dist=pft_back_tracking_dist,
        pft_front_tracking_dist=pft_front_tracking_dist,
        particle_count=particle_count,
        pft_max_trial=pft_max_trial,
        return_all=return_all,
        random_seed=random_seed,
        initial_directions=seed_directions,
        save_seeds=save_seeds,
        min_wm_pve_before_stopping=min_wm_pve_before_stopping,
        unidirectional=unidirectional,
        randomize_forward_direction=randomize_forward_direction,
    )



import os.path as osp
from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.testing import assert_true


def test_read_bvals_bvecs():
    fimg, fbvals, fbvecs = get_fnames(name="small_101D")
    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
    gt = gradient_table(bvals, bvecs=bvecs)
    npt.assert_array_equal(bvals, gt.bvals)
    npt.assert_array_equal(bvecs, gt.bvecs)

    # None should also work as an input:
    bvals_none, bvecs_none = read_bvals_bvecs(None, fbvecs)
    npt.assert_array_equal(bvecs_none, gt.bvecs)
    bvals_none, bvecs_none = read_bvals_bvecs(fbvals, None)
    npt.assert_array_equal(bvals_none, gt.bvals)

    # Test for error raising with unknown file formats:
    nan_fbvecs = osp.splitext(fbvecs)[0] + ".nan"  # Nonsense extension
    npt.assert_raises(ValueError, read_bvals_bvecs, fbvals, nan_fbvecs)
    npt.assert_raises(ValueError, read_bvals_bvecs, bvals, nan_fbvecs)
    npt.assert_raises(ValueError, read_bvals_bvecs, fbvals, bvecs)

    # Test for error raising with incorrect file-contents:

    with TemporaryDirectory() as tmpdir:
        # These bvecs only have two rows/columns:
        new_bvecs1 = bvecs[:, :2]
        # Make a temporary file
        fname = "test_bv_file1.txt"
        with open(pjoin(tmpdir, fname), "wt") as bv_file1:
            # And fill it with these 2-columned bvecs:
            for x in range(new_bvecs1.shape[0]):
                bv_file1.write(f"{new_bvecs1[x][0]} {new_bvecs1[x][1]}\n")
        npt.assert_raises(OSError, read_bvals_bvecs, fbvals, fname)

        # These bvecs are saved as one long array:
        fname = "test_bv_file2.npy"
        new_bvecs2 = np.ravel(bvecs)
        with open(pjoin(tmpdir, fname), "w") as bv_file2:
            np.save(bv_file2.name, new_bvecs2)
        npt.assert_raises(OSError, read_bvals_bvecs, fbvals, fname)

        # There are less bvecs than bvals:
        fname = "test_bv_file3.txt"
        new_bvecs3 = bvecs[:-1, :]
        with open(pjoin(tmpdir, fname), "w") as bv_file3:
            np.savetxt(bv_file3.name, new_bvecs3)
        npt.assert_raises(OSError, read_bvals_bvecs, fbvals, fname)

        # You entered the bvecs on both sides:
        npt.assert_raises(OSError, read_bvals_bvecs, fbvecs, fbvecs)

        # All possible delimiters should work
        bv_file4 = "test_space.txt"
        with open(pjoin(tmpdir, bv_file4), "w") as f:
            f.write("66 55 33")
        bvals_1, _ = read_bvals_bvecs(pjoin(tmpdir, bv_file4), "")

        bv_file5 = "test_coma.txt"
        with open(pjoin(tmpdir, bv_file5), "w") as f:
            f.write("66, 55, 33")
        bvals_2, _ = read_bvals_bvecs(pjoin(tmpdir, bv_file5), "")

        bv_file6 = "test_tabs.txt"
        with open(pjoin(tmpdir, bv_file6), "w") as f:
            f.write("66 \t 55 \t 33")
        bvals_3, _ = read_bvals_bvecs(pjoin(tmpdir, bv_file6), "")

        ans = np.array([66.0, 55.0, 33.0])
        npt.assert_array_equal(ans, bvals_1)
        npt.assert_array_equal(ans, bvals_2)
        npt.assert_array_equal(ans, bvals_3)

        bv_file7 = "test_space_2.txt"
        with open(pjoin(tmpdir, bv_file7), "w") as f:
            f.write("66 55 33\n45 34 21\n55 32 65\n")
        _, bvecs_1 = read_bvals_bvecs("", pjoin(tmpdir, bv_file7))

        bv_file8 = "test_coma_2.txt"
        with open(pjoin(tmpdir, bv_file8), "w") as f:
            f.write("66, 55, 33\n45, 34, 21 \n 55, 32, 65\n")
        _, bvecs_2 = read_bvals_bvecs("", pjoin(tmpdir, bv_file8))

        bv_file9 = "test_tabs_2.txt"
        with open(pjoin(tmpdir, bv_file9), "w") as f:
            f.write("66 \t 55 \t 33\n45 \t 34 \t 21\n55 \t 32 \t 65\n")
        _, bvecs_3 = read_bvals_bvecs("", pjoin(tmpdir, bv_file9))

        bv_file10 = "test_multiple_space.txt"
        with open(pjoin(tmpdir, bv_file10), "w") as f:
            f.write("66   55   33\n45,   34,   21 \n 55,   32,     65\n")
        _, bvecs_4 = read_bvals_bvecs("", pjoin(tmpdir, bv_file10))

        ans = np.array([[66.0, 55.0, 33.0], [45.0, 34.0, 21.0], [55.0, 32.0, 65.0]])
        npt.assert_array_equal(ans, bvecs_1)
        npt.assert_array_equal(ans, bvecs_2)
        npt.assert_array_equal(ans, bvecs_3)
        npt.assert_array_equal(ans, bvecs_4)

        bv_two_volume = "bv_two_volume.txt"
        with open(pjoin(tmpdir, bv_two_volume), "w") as f:
            f.write("0 0 \n0 0 \n0 0")
        bval_two_volume = "bval_two_volume.txt"
        with open(pjoin(tmpdir, bval_two_volume), "w") as f:
            f.write("0\n0\n")
        bval_5, bvecs_5 = read_bvals_bvecs(
            pjoin(tmpdir, bval_two_volume), pjoin(tmpdir, bv_two_volume)
        )
        npt.assert_array_equal(bvecs_5, np.zeros((2, 3)))
        npt.assert_array_equal(bval_5, np.zeros(2))

        bv_single_volume = "test_single_volume.txt"
        with open(pjoin(tmpdir, bv_single_volume), "w") as f:
            f.write("0 \n0 \n0 ")
        bval_single_volume = "test_single_volume_2.txt"
        with open(pjoin(tmpdir, bval_single_volume), "w") as f:
            f.write("0\n")

        with warnings.catch_warnings(record=True) as w:
            bval_5, bvecs_5 = read_bvals_bvecs(
                pjoin(tmpdir, bval_single_volume), pjoin(tmpdir, bv_single_volume)
            )
            npt.assert_array_equal(bvecs_5, np.zeros((1, 3)))
            npt.assert_array_equal(bval_5, np.zeros(1))
            assert_true(len(w) == 1)
            assert_true(issubclass(w[0].category, UserWarning))
            assert_true("Detected only 1 direction on" in str(w[0].message))



from numpy.testing import assert_equal

from dipy.core.graph import Graph


def test_graph():
    g = Graph()

    g.add_node("a", attr=5)
    g.add_node("b", attr=6)
    g.add_node("c", attr=10)
    g.add_node("d", attr=11)

    g.add_edge("a", "b")
    g.add_edge("b", "c")
    g.add_edge("c", "d")
    g.add_edge("b", "d")

    print("Nodes")
    print(g.node)
    print("Successors")
    print(g.succ)
    print("Predecessors")
    print(g.pred)
    print("Paths above d")
    print(g.up("d"))
    print("Paths below a")
    print(g.down("a"))
    print("Shortest path above d")
    print(g.up_short("d"))
    print("Shortest path below a")
    print(g.down_short("a"))
    print("Deleting node b")
    # g.del_node_and_edges('b')
    g.del_node("b")
    print("Nodes")
    print(g.node)
    print("Successors")
    print(g.succ)
    print("Predecessors")
    print(g.pred)

    assert_equal(len(g.node), 3)
    assert_equal(len(g.succ), 3)
    assert_equal(len(g.pred), 3)
